"""MCP tools for BPA operations.

This module provides tools for interacting with the BPA API through MCP.
"""

from mcp_eregistrations_bpa.tools.actions import (
    componentaction_delete,
    componentaction_get,
    componentaction_get_by_component,
    componentaction_get_system_actions,
    componentaction_save,
    componentaction_update,
    register_action_tools,
)
from mcp_eregistrations_bpa.tools.analysis import (
    analyze_service,
    register_analysis_tools,
)
from mcp_eregistrations_bpa.tools.audit import (
    audit_get,
    audit_list,
    register_audit_tools,
)
from mcp_eregistrations_bpa.tools.behaviours import (
    componentbehaviour_get,
    componentbehaviour_get_by_component,
    componentbehaviour_list,
    componentbehaviour_repair_metadata,
    effect_create,
    effect_delete,
    register_behaviour_tools,
)
from mcp_eregistrations_bpa.tools.bot_mappings import (
    bot_clone,
    bot_history_list,
    bot_history_revert,
    bot_input_mapping_create,
    bot_input_mapping_delete,
    bot_input_mapping_list,
    bot_input_mapping_save_all,
    bot_input_mapping_update,
    bot_input_visibility_update,
    bot_mapping_summary,
    bot_output_mapping_create,
    bot_output_mapping_delete,
    bot_output_mapping_list,
    bot_output_mapping_save_all,
    bot_output_mapping_update,
    bot_output_visibility_update,
    bot_suggest_mappings,
    register_bot_mapping_tools,
)
from mcp_eregistrations_bpa.tools.bots import (
    bot_create,
    bot_delete,
    bot_get,
    bot_list,
    bot_update,
    bot_upgrade_version,
    bot_validate,
    muleservice_discover,
    register_bot_tools,
)
from mcp_eregistrations_bpa.tools.classifications import (
    classification_create,
    classification_export_csv,
    classification_get,
    classification_list,
    classification_search,
    classification_update,
    register_classification_tools,
)
from mcp_eregistrations_bpa.tools.costs import (
    cost_create_fixed,
    cost_create_formula,
    cost_delete,
    cost_update,
    register_cost_tools,
)
from mcp_eregistrations_bpa.tools.debugger import (
    debug_fix,
    debug_fix_batch,
    debug_group_issues,
    debug_investigate,
    debug_plan,
    debug_scan,
    debug_verify,
    register_debug_tools,
)
from mcp_eregistrations_bpa.tools.determinants import (
    booleandeterminant_create,
    classificationdeterminant_create,
    datedeterminant_create,
    determinant_delete,
    determinant_get,
    determinant_list,
    determinant_search,
    griddeterminant_create,
    numericdeterminant_create,
    register_determinant_tools,
    selectdeterminant_create,
    textdeterminant_create,
    textdeterminant_update,
)
from mcp_eregistrations_bpa.tools.document_requirements import (
    documentrequirement_create,
    documentrequirement_delete,
    documentrequirement_list,
    documentrequirement_update,
    register_document_requirement_tools,
    requirement_list,
)
from mcp_eregistrations_bpa.tools.export import (
    register_export_tools,
    service_copy,
    service_export_raw,
    service_to_yaml,
)
from mcp_eregistrations_bpa.tools.external_services import (
    muleservice_get,
    muleservice_list,
    register_external_service_tools,
)
from mcp_eregistrations_bpa.tools.fields import (
    field_get,
    field_list,
    register_field_tools,
)
from mcp_eregistrations_bpa.tools.form_cache_tools import (
    form_query,
    register_form_cache_tools,
)
from mcp_eregistrations_bpa.tools.formio_helpers import (
    build_checkbox,
    build_columns,
    build_number,
    build_panel,
    build_select,
    build_textarea,
    build_textfield,
)
from mcp_eregistrations_bpa.tools.forms import (
    form_component_add,
    form_component_get,
    form_component_move,
    form_component_remove,
    form_component_update,
    form_get,
    form_patch,
    form_update,
    register_form_tools,
)
from mcp_eregistrations_bpa.tools.instances import (
    instance_add,
    instance_list,
    instance_remove,
    register_instance_tools,
)
from mcp_eregistrations_bpa.tools.messages import (
    message_create,
    message_delete,
    message_get,
    message_list,
    message_update,
    register_message_tools,
)
from mcp_eregistrations_bpa.tools.notifications import (
    notification_create,
    notification_list,
    register_notification_tools,
)
from mcp_eregistrations_bpa.tools.print_documents import (
    print_document_component_add,
    print_document_component_get,
    print_document_component_move,
    print_document_component_remove,
    print_document_component_update,
    print_document_create,
    print_document_delete,
    print_document_get,
    print_document_history,
    print_document_list,
    print_document_patch,
    print_document_revert,
    print_document_sort,
    print_document_templates,
    print_document_update,
    register_print_document_tools,
)
from mcp_eregistrations_bpa.tools.registration_institutions import (
    institution_create,
    institution_discover,
    register_registration_institution_tools,
    registrationinstitution_create,
    registrationinstitution_delete,
    registrationinstitution_get,
    registrationinstitution_list,
    registrationinstitution_list_by_institution,
)
from mcp_eregistrations_bpa.tools.registrations import (
    register_registration_tools,
    registration_activate,
    registration_create,
    registration_delete,
    registration_get,
    registration_list,
    serviceregistration_link,
)
from mcp_eregistrations_bpa.tools.role_forms import (
    register_role_form_tools,
    role_form_component_add,
    role_form_component_remove,
    role_form_component_update,
    role_form_patch,
)
from mcp_eregistrations_bpa.tools.role_status import (
    register_role_status_tools,
    rolestatus_create,
    rolestatus_delete,
    rolestatus_get,
    rolestatus_update,
)
from mcp_eregistrations_bpa.tools.role_units import (
    register_role_unit_tools,
    roleunit_create,
    roleunit_delete,
    roleunit_get,
    roleunit_list,
)
from mcp_eregistrations_bpa.tools.roles import (
    register_role_tools,
    role_create,
    role_delete,
    role_get,
    role_list,
    role_update,
    roleinstitution_create,
    roleregistration_create,
)
from mcp_eregistrations_bpa.tools.rollback import (
    register_rollback_tools,
    rollback,
    rollback_cleanup,
    rollback_history,
)
from mcp_eregistrations_bpa.tools.service_insights import (
    register_service_insights_tools,
    service_branches,
    workflow_graph,
)
from mcp_eregistrations_bpa.tools.service_map import (
    register_service_map_tools,
    service_ai_map,
)
from mcp_eregistrations_bpa.tools.services import (
    register_service_tools,
    service_activate,
    service_create,
    service_get,
    service_list,
    service_publish,
    service_update,
)
from mcp_eregistrations_bpa.tools.workflows import (
    register_workflow_tools,
    workflow_cancel,
    workflow_chain,
    workflow_confirm,
    workflow_continue,
    workflow_describe,
    workflow_execute,
    workflow_list,
    workflow_retry,
    workflow_rollback,
    workflow_search,
    workflow_start_interactive,
    workflow_status,
)

__all__ = [
    # Service tools
    "service_list",
    "service_get",
    "service_create",
    "service_update",
    "service_publish",
    "service_activate",
    "register_service_tools",
    # Registration tools
    "registration_list",
    "registration_get",
    "registration_create",
    "registration_delete",
    "registration_activate",
    "serviceregistration_link",
    "register_registration_tools",
    # Registration institution tools
    "registrationinstitution_list",
    "registrationinstitution_get",
    "registrationinstitution_create",
    "registrationinstitution_delete",
    "registrationinstitution_list_by_institution",
    "institution_discover",
    "institution_create",
    "register_registration_institution_tools",
    # Field tools
    "field_list",
    "field_get",
    "register_field_tools",
    # Form tools
    "form_get",
    "form_component_get",
    "form_component_add",
    "form_component_update",
    "form_component_remove",
    "form_component_move",
    "form_update",
    "form_patch",
    "register_form_tools",
    # Form cache tools
    "form_query",
    "register_form_cache_tools",
    # Form.io component builders
    "build_textfield",
    "build_number",
    "build_select",
    "build_checkbox",
    "build_textarea",
    "build_panel",
    "build_columns",
    # Determinant tools
    "determinant_list",
    "determinant_get",
    "determinant_search",
    "determinant_delete",
    "textdeterminant_create",
    "textdeterminant_update",
    "selectdeterminant_create",
    "numericdeterminant_create",
    "booleandeterminant_create",
    "datedeterminant_create",
    "classificationdeterminant_create",
    "griddeterminant_create",
    "register_determinant_tools",
    # Behaviour/Effect tools
    "componentbehaviour_list",
    "componentbehaviour_get",
    "componentbehaviour_get_by_component",
    "componentbehaviour_repair_metadata",
    "effect_create",
    "effect_delete",
    "register_behaviour_tools",
    # Component action tools
    "componentaction_get",
    "componentaction_get_by_component",
    "componentaction_get_system_actions",
    "componentaction_save",
    "componentaction_update",
    "componentaction_delete",
    "register_action_tools",
    # Bot tools
    "bot_list",
    "bot_get",
    "bot_create",
    "bot_update",
    "bot_delete",
    "bot_validate",
    "bot_upgrade_version",
    "muleservice_discover",
    "register_bot_tools",
    # Bot mapping tools
    "bot_input_mapping_list",
    "bot_output_mapping_list",
    "bot_input_mapping_create",
    "bot_input_mapping_update",
    "bot_input_mapping_delete",
    "bot_output_mapping_create",
    "bot_output_mapping_update",
    "bot_output_mapping_delete",
    "bot_input_mapping_save_all",
    "bot_output_mapping_save_all",
    "bot_clone",
    "bot_history_list",
    "bot_history_revert",
    "bot_input_visibility_update",
    "bot_output_visibility_update",
    "bot_mapping_summary",
    "bot_suggest_mappings",
    "register_bot_mapping_tools",
    # External service tools
    "muleservice_list",
    "muleservice_get",
    "register_external_service_tools",
    # Classification tools
    "classification_list",
    "classification_get",
    "classification_search",
    "classification_create",
    "classification_update",
    "classification_export_csv",
    "register_classification_tools",
    # Notification tools
    "notification_list",
    "notification_create",
    "register_notification_tools",
    # Role tools
    "role_list",
    "role_get",
    "role_create",
    "role_update",
    "role_delete",
    "roleinstitution_create",
    "roleregistration_create",
    "register_role_tools",
    # Role form tools
    "role_form_component_update",
    "role_form_component_add",
    "role_form_component_remove",
    "role_form_patch",
    "register_role_form_tools",
    # Role status tools
    "rolestatus_get",
    "rolestatus_create",
    "rolestatus_update",
    "rolestatus_delete",
    "register_role_status_tools",
    # Role unit tools
    "roleunit_list",
    "roleunit_get",
    "roleunit_create",
    "roleunit_delete",
    "register_role_unit_tools",
    # Message tools
    "message_list",
    "message_get",
    "message_create",
    "message_update",
    "message_delete",
    "register_message_tools",
    # Analysis tools
    "analyze_service",
    "register_analysis_tools",
    # Document requirement tools
    "requirement_list",
    "documentrequirement_list",
    "documentrequirement_create",
    "documentrequirement_update",
    "documentrequirement_delete",
    "register_document_requirement_tools",
    # Cost tools
    "cost_create_fixed",
    "cost_create_formula",
    "cost_update",
    "cost_delete",
    "register_cost_tools",
    # Audit tools
    "audit_list",
    "audit_get",
    "register_audit_tools",
    # Rollback tools
    "rollback",
    "rollback_history",
    "rollback_cleanup",
    "register_rollback_tools",
    # Instance management tools
    "instance_list",
    "instance_add",
    "instance_remove",
    "register_instance_tools",
    # Export tools
    "service_export_raw",
    "service_to_yaml",
    "service_copy",
    "register_export_tools",
    # Workflow orchestration tools
    "workflow_list",
    "workflow_describe",
    "workflow_search",
    "workflow_execute",
    "workflow_status",
    "workflow_cancel",
    "workflow_retry",
    "workflow_rollback",
    "workflow_chain",
    "workflow_start_interactive",
    "workflow_continue",
    "workflow_confirm",
    "register_workflow_tools",
    # Print document tools
    "print_document_list",
    "print_document_get",
    "print_document_create",
    "print_document_update",
    "print_document_delete",
    "print_document_component_get",
    "print_document_component_add",
    "print_document_component_update",
    "print_document_component_remove",
    "print_document_component_move",
    "print_document_sort",
    "print_document_templates",
    "print_document_history",
    "print_document_revert",
    "print_document_patch",
    "register_print_document_tools",
    # Service insights tools
    "workflow_graph",
    "service_branches",
    "register_service_insights_tools",
    # Service map tools
    "service_ai_map",
    "register_service_map_tools",
    # Debug tools
    "debug_scan",
    "debug_investigate",
    "debug_fix",
    "debug_fix_batch",
    "debug_group_issues",
    "debug_plan",
    "debug_verify",
    "register_debug_tools",
]



"""MCP tools for BPA external service (Mule) operations.

This module provides tools for listing and retrieving external services (GDB APIs)
available in the BPA platform. These are used by bots to connect to external data
sources.

API Endpoints used:
- GET /mule/services - List all available external services
- GET /mule/services/{service_id} - Get external service by ID (base64 encoded)
"""

from __future__ import annotations

import base64
from typing import Any

from fastmcp.exceptions import ToolError

from mcp_eregistrations_bpa.bpa_client import BPAClient
from mcp_eregistrations_bpa.bpa_client.errors import (
    BPAClientError,
    BPANotFoundError,
    translate_error,
)
from mcp_eregistrations_bpa.tools.large_response import large_response_handler

__all__ = [
    "muleservice_list",
    "muleservice_get",
    "register_external_service_tools",
]


def _transform_muleservice_response(data: dict[str, Any]) -> dict[str, Any]:
    """Transform mule service API response from camelCase to snake_case.

    Args:
        data: Raw API response with camelCase keys.

    Returns:
        dict: Transformed response with snake_case keys.
    """
    return {
        "id": data.get("id"),
        "name": data.get("name"),
        "type": data.get("type"),
        "db_name": data.get("dbName"),
        "base_url": data.get("baseUrl"),
        "method": data.get("method"),
        "inputs": data.get("inputs", []),
        "outputs": data.get("outputs", []),
    }


def _transform_muleservice_list_item(data: dict[str, Any]) -> dict[str, Any]:
    """Transform mule service list item (minimal fields).

    Args:
        data: Raw API response with camelCase keys.

    Returns:
        dict: Transformed response with snake_case keys.
    """
    return {
        "id": data.get("id"),
        "name": data.get("name"),
        "type": data.get("type"),
        "db_name": data.get("dbName"),
    }


@large_response_handler(
    threshold_bytes=50 * 1024,
    navigation={
        "list_all": "jq '.services'",
        "gdb_only": "jq '.services[] | select(.id | startswith(\"GDB.\"))'",
        "find_by_name": "jq '.services[] | select(.name | contains(\"search\"))'",
    },
)
async def muleservice_list(
    service_type: str | None = None,
    name_filter: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """List available external services (GDB, APIs).

    Args:
        service_type: Filter by type prefix (e.g., "GDB").
        name_filter: Filter by name substring (case-insensitive).
        instance: Instance profile name (from instance_list).

    Returns:
        dict with services (list of id, name, type, db_name), total.
    """
    try:
        async with BPAClient.for_instance(instance) as client:
            result = await client.get(
                "/mule/services",
                params={"withoutdata": "true"},
                resource_type="mule_service",
            )

            # Handle both list and dict responses
            if isinstance(result, list):
                services_raw = result
            else:
                services_raw = result.get("services", result.get("data", []))

            # Apply filters
            services = []
            for svc in services_raw:
                svc_id = svc.get("id", "")
                svc_name = svc.get("name", "")

                # Filter by service type prefix
                if service_type:
                    if not svc_id.upper().startswith(service_type.upper()):
                        continue

                # Filter by name substring
                if name_filter:
                    if name_filter.lower() not in svc_name.lower():
                        continue

                services.append(_transform_muleservice_list_item(svc))

            return {
                "services": services,
                "total": len(services),
                "filters_applied": {
                    "service_type": service_type,
                    "name_filter": name_filter,
                },
            }

    except BPAClientError as e:
        raise translate_error(e, resource_type="mule_service")


async def muleservice_get(
    service_id: str, instance: str | None = None
) -> dict[str, Any]:
    """Get external service details including input/output mappings.

    Args:
        service_id: Service ID (e.g., GDB.GDB-NIPC REGISTRY(2.9)-read).
        instance: Instance profile name (from instance_list).

    Returns:
        dict with id, name, type, db_name, base_url, inputs, outputs.
    """
    if not service_id:
        raise ToolError(
            "Cannot get mule service: 'service_id' is required. "
            "Use 'muleservice_list' to see available services."
        )

    # Service IDs need to be base64 encoded for the API
    encoded_id = base64.b64encode(service_id.encode()).decode()

    try:
        async with BPAClient.for_instance(instance) as client:
            try:
                service_data = await client.get(
                    "/mule/services/{service_id}",
                    path_params={"service_id": encoded_id},
                    resource_type="mule_service",
                    resource_id=service_id,
                )
            except BPANotFoundError:
                raise ToolError(
                    f"Mule service '{service_id}' not found. "
                    "Use 'muleservice_list' to see available services."
                )

        return _transform_muleservice_response(service_data)

    except ToolError:
        raise
    except BPAClientError as e:
        raise translate_error(e, resource_type="mule_service", resource_id=service_id)


def register_external_service_tools(mcp: Any) -> None:
    """Register external service tools with the MCP server.

    Args:
        mcp: The FastMCP server instance.
    """
    from mcp_eregistrations_bpa.tools.annotations import READ

    # Read operations
    mcp.tool(annotations=READ)(muleservice_list)
    mcp.tool(annotations=READ)(muleservice_get)



"""Database schema creation and migration system.

Provides versioned schema migrations for audit_logs and rollback_states tables.
"""

from pathlib import Path
from typing import Any

import aiosqlite

from mcp_eregistrations_bpa.db.connection import get_connection, get_db_path

# Migration definitions: (version, sql_statements)
# Each migration is a tuple of (version_number, sql_script)
# Migrations are applied in order, only if version > current_version
MIGRATIONS: list[tuple[int, str]] = [
    (
        1,
        """
        -- Version tracking table
        CREATE TABLE IF NOT EXISTS schema_version (
            version INTEGER PRIMARY KEY,
            applied_at TEXT NOT NULL DEFAULT (datetime('now'))
        );

        -- Audit logs table (NFR5: tamper-evident, append-only)
        CREATE TABLE IF NOT EXISTS audit_logs (
            id TEXT PRIMARY KEY,
            timestamp TEXT NOT NULL,
            user_email TEXT NOT NULL,
            operation_type TEXT NOT NULL,
            object_type TEXT NOT NULL,
            object_id TEXT,
            params TEXT NOT NULL,
            status TEXT NOT NULL DEFAULT 'pending',
            result TEXT,
            rollback_state_id TEXT
        );

        -- Rollback states table
        CREATE TABLE IF NOT EXISTS rollback_states (
            id TEXT PRIMARY KEY,
            audit_log_id TEXT NOT NULL REFERENCES audit_logs(id),
            object_type TEXT NOT NULL,
            object_id TEXT NOT NULL,
            previous_state TEXT NOT NULL,
            created_at TEXT NOT NULL DEFAULT (datetime('now'))
        );

        -- Indexes for common queries
        CREATE INDEX IF NOT EXISTS idx_audit_logs_timestamp
            ON audit_logs(timestamp);
        CREATE INDEX IF NOT EXISTS idx_audit_logs_status
            ON audit_logs(status);
        CREATE INDEX IF NOT EXISTS idx_audit_logs_object_type
            ON audit_logs(object_type);
        CREATE INDEX IF NOT EXISTS idx_audit_logs_object_id
            ON audit_logs(object_id);
        CREATE INDEX IF NOT EXISTS idx_rollback_states_audit_log_id
            ON rollback_states(audit_log_id);
        CREATE INDEX IF NOT EXISTS idx_rollback_states_object
            ON rollback_states(object_type, object_id);
        """,
    ),
    (
        2,
        """
        -- Add index on user_email for filtering by user
        CREATE INDEX IF NOT EXISTS idx_audit_logs_user_email
            ON audit_logs(user_email);
        """,
    ),
    (
        3,
        """
        -- Form cache: local SQLite cache for BPA form schemas
        CREATE TABLE IF NOT EXISTS form_cache (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            service_id TEXT NOT NULL,
            form_type TEXT NOT NULL,
            form_data TEXT NOT NULL,
            form_schema TEXT NOT NULL,
            cached_at TEXT NOT NULL DEFAULT (datetime('now')),
            expires_at TEXT NOT NULL,
            UNIQUE(service_id, form_type)
        );

        CREATE INDEX IF NOT EXISTS idx_form_cache_lookup
            ON form_cache(service_id, form_type);
        CREATE INDEX IF NOT EXISTS idx_form_cache_expires
            ON form_cache(expires_at);
        """,
    ),
]


async def get_schema_version(conn: aiosqlite.Connection) -> int:
    """Get the current schema version from the database.

    Args:
        conn: Active database connection.

    Returns:
        Current schema version number, or 0 if no version table exists.
    """
    try:
        cursor = await conn.execute("SELECT MAX(version) FROM schema_version")
        row = await cursor.fetchone()
        if row and row[0] is not None:
            version: int = row[0]
            return version
        return 0
    except aiosqlite.OperationalError:
        # Table doesn't exist yet
        return 0


async def apply_migrations(conn: aiosqlite.Connection) -> int:
    """Apply pending migrations in order.

    Args:
        conn: Active database connection.

    Returns:
        Number of migrations applied.

    Raises:
        aiosqlite.Error: If migration fails.
    """
    current_version = await get_schema_version(conn)
    migrations_applied = 0

    for version, sql in MIGRATIONS:
        if version > current_version:
            # Execute the migration script
            await conn.executescript(sql)
            # Record the applied migration
            await conn.execute(
                "INSERT INTO schema_version (version) VALUES (?)",
                (version,),
            )
            await conn.commit()
            migrations_applied += 1

    return migrations_applied


async def initialize_database(db_path: Path | None = None) -> dict[str, Any]:
    """Initialize the database with all required tables and indexes.

    This is the main entry point for database setup. It should be called
    during MCP server startup.

    Args:
        db_path: Optional path to database. Defaults to get_db_path() if not provided.

    Returns:
        Dictionary with initialization results:
        - db_path: Path to the database file
        - schema_version: Current schema version after initialization
        - migrations_applied: Number of migrations applied
        - tables_created: List of tables in the database

    Example:
        result = await initialize_database()
        print(f"Database at {result['db_path']}")
        print(f"Schema version: {result['schema_version']}")
    """
    if db_path is None:
        db_path = get_db_path()

    async with get_connection(db_path) as conn:
        migrations_applied = await apply_migrations(conn)
        schema_version = await get_schema_version(conn)

        # Get list of tables
        cursor = await conn.execute(
            "SELECT name FROM sqlite_master WHERE type='table' ORDER BY name"
        )
        rows = await cursor.fetchall()
        tables = [row[0] for row in rows]

    return {
        "db_path": str(db_path),
        "schema_version": schema_version,
        "migrations_applied": migrations_applied,
        "tables_created": tables,
    }



from fastmcp import FastMCP

mcp = FastMCP(
    "mcp-eregistrations-smartlink",
    instructions=(
        "SmartLink debugging tools for eRegistrations integration flows. "
        "Start with smartlink_audit_get to inspect a failing request, then use "
        "smartlink_flow_get / smartlink_action_get to understand the configuration. "
        "Use smartlink_flow_update / smartlink_action_update to fix broken config.\n\n"
        "AUTHENTICATION: If a tool returns an auth error, call smartlink_auth_login "
        "with the user's credentials. Credentials are shared across all eRegistrations "
        "MCP servers.\n\n"
        "NOTE: The authenticated user must have the sysadmin role to access audit data."
    ),
)

import mcp_eregistrations_smartlink.tools.actions  # noqa: F401, E402
import mcp_eregistrations_smartlink.tools.audits  # noqa: F401, E402
import mcp_eregistrations_smartlink.tools.flows  # noqa: F401, E402
import mcp_eregistrations_smartlink.tools.instances  # noqa: F401, E402
import mcp_eregistrations_smartlink.tools.system  # noqa: F401, E402


def main() -> None:
    """Entry point for mcp-eregistrations-smartlink."""
    mcp.run()



"""Form.io schema manipulation helpers.

This module provides utility functions for manipulating Form.io component trees.
It supports finding, inserting, removing, and updating components within the
nested structure (panels, columns, tabs, etc.).

All functions operate on copies to avoid mutating original data.

Container Type Hierarchy
------------------------
Form.io uses container components to organize form layouts. Each container type
stores its children in a specific property:

+-------------+---------------------------+----------------------------------------+
| Type        | Children Accessor         | Notes                                  |
+-------------+---------------------------+----------------------------------------+
| tabs        | components[]              | Tab panes have null type; each pane    |
|             |                           | contains its own components[]          |
+-------------+---------------------------+----------------------------------------+
| panel       | components[]              | Standard collapsible container         |
+-------------+---------------------------+----------------------------------------+
| columns     | columns[].components[]    | 2-level nesting: columns array with    |
|             |                           | each column having a components array  |
+-------------+---------------------------+----------------------------------------+
| fieldset    | components[]              | HTML fieldset grouping                 |
+-------------+---------------------------+----------------------------------------+
| editgrid    | components[]              | Repeatable rows with inline editing    |
+-------------+---------------------------+----------------------------------------+
| datagrid    | components[]              | Repeatable rows in table format        |
+-------------+---------------------------+----------------------------------------+
| table       | rows[][]                  | HTML table structure: rows array       |
|             |                           | containing arrays of cells             |
+-------------+---------------------------+----------------------------------------+
| well        | components[]              | Visual grouping container              |
+-------------+---------------------------+----------------------------------------+
| container   | components[]              | Data grouping without visual styling   |
+-------------+---------------------------+----------------------------------------+
| dropdown    | components[]              | Collapsible dropdown container         |
+-------------+---------------------------+----------------------------------------+

Path Examples
-------------
When traversing nested components, paths represent the hierarchy from root to
the target component:

    # Example: component inside tabs > tab_pane > panel
    path = ["applicantWelcome", "applicantWelcomepersonalInformation", "applicantBlock"]
    # Meaning: tabs("applicantWelcome") > panel("applicantWelcomepersonalInformation")
    #          > panel("applicantBlock") > [target component]

    # Example: component inside columns
    path = ["myPanel", "twoColumnLayout"]
    # Meaning: panel("myPanel") > columns("twoColumnLayout") > [target component]

The path array is returned by find_component() and used in API responses
to help identify the nesting structure of each component.
"""

from __future__ import annotations

import copy
import re
from typing import Any

# Container component types that can hold nested components
CONTAINER_TYPES = {
    "panel",
    "columns",
    "tabs",
    "tab",
    "well",
    "fieldset",
    "container",
    "editgrid",
    "datagrid",
    "table",
    "dropdown",
}

# Container types with special child structures (not just "components" array)
# These require additional parameters (column_index, row_index, etc.) when inserting
COLUMN_BASED_CONTAINERS = {"columns"}
# TODO: Table support - tables use rows[i].components[] structure similar to columns
# Implementation deferred until table editing is needed
TABLE_BASED_CONTAINERS = {"table"}

# Maximum recursion depth for tree traversal (prevent stack overflow)
MAX_RECURSION_DEPTH = 100

# BPA-specific default properties for new components
BPA_COMPONENT_DEFAULTS = {
    "registrations": {},
    "determinantIds": [],
    "componentActionId": "",
    "componentFormulaId": "",
    "behaviourId": "",
    "componentValidationId": "",
    "version": "201905",
    "input": True,
    "tableView": True,
}


def collect_registration_ids(
    components: list[dict[str, Any]],
    _depth: int = 0,
) -> dict[str, bool]:
    """Collect all registration IDs from existing components.

    Walks the component tree and builds a union of all registration assignments.
    Used to auto-assign registrations to new components (matching frontend behavior).

    Args:
        components: Form.io components list.

    Returns:
        Dict of {registration_id: True} for all registrations found.
    """
    if _depth > MAX_RECURSION_DEPTH:
        return {}

    result: dict[str, bool] = {}
    for comp in components:
        if not isinstance(comp, dict):
            continue
        regs = comp.get("registrations")
        if isinstance(regs, dict):
            for reg_id, val in regs.items():
                if val:
                    result[reg_id] = True
        # Recurse into children
        for child_key in ("components", "columns", "rows"):
            children = comp.get(child_key)
            if isinstance(children, list):
                if child_key == "components":
                    # Direct children are components — recurse into them
                    result.update(collect_registration_ids(children, _depth + 1))
                else:
                    # columns/rows: each child is a wrapper with "components"
                    for child in children:
                        if isinstance(child, dict):
                            child_comps = child.get("components", [])
                            if isinstance(child_comps, list):
                                result.update(
                                    collect_registration_ids(child_comps, _depth + 1)
                                )
    return result


def find_component(
    components: list[dict[str, Any]],
    key: str,
    path: list[str] | None = None,
    _depth: int = 0,
) -> tuple[dict[str, Any], list[str]] | None:
    """Find a component by key in the component tree.

    Searches recursively through nested components (panels, columns, tabs).

    Args:
        components: List of Form.io components to search.
        key: The component key to find.
        path: Current path (used internally for recursion).
        _depth: Current recursion depth (internal use).

    Returns:
        Tuple of (component, path) if found, None otherwise.
        The path is a list of parent keys leading to the component.

    Raises:
        RecursionError: If nesting exceeds MAX_RECURSION_DEPTH.

    Example:
        >>> components = [{"key": "panel1", "type": "panel", "components": [
        ...     {"key": "name", "type": "textfield"}
        ... ]}]
        >>> result = find_component(components, "name")
        >>> result[0]["key"]
        'name'
        >>> result[1]
        ['panel1']
    """
    if _depth > MAX_RECURSION_DEPTH:
        raise RecursionError(
            f"Component tree exceeds maximum depth of {MAX_RECURSION_DEPTH}"
        )

    if path is None:
        path = []

    for comp in components:
        if not isinstance(comp, dict):
            continue

        comp_key = comp.get("key")
        if comp_key == key:
            return (comp, path)

        new_path = path + [comp_key] if comp_key else path

        # Search in nested components
        nested = comp.get("components", [])
        if isinstance(nested, list) and nested:
            result = find_component(nested, key, new_path, _depth + 1)
            if result:
                return result

        # Search in columns
        columns = comp.get("columns", [])
        if isinstance(columns, list):
            for col in columns:
                if isinstance(col, dict):
                    col_components = col.get("components", [])
                    if isinstance(col_components, list) and col_components:
                        result = find_component(
                            col_components, key, new_path, _depth + 1
                        )
                        if result:
                            return result

        # Search in tabs
        tabs = comp.get("tabs", [])
        if isinstance(tabs, list):
            for tab in tabs:
                if isinstance(tab, dict):
                    tab_components = tab.get("components", [])
                    if isinstance(tab_components, list) and tab_components:
                        result = find_component(
                            tab_components, key, new_path, _depth + 1
                        )
                        if result:
                            return result

        # Search in table rows (rows[][] structure)
        rows = comp.get("rows", [])
        if isinstance(rows, list):
            for row in rows:
                if isinstance(row, list):
                    for cell in row:
                        if isinstance(cell, dict):
                            cell_components = cell.get("components", [])
                            if isinstance(cell_components, list) and cell_components:
                                result = find_component(
                                    cell_components, key, new_path, _depth + 1
                                )
                                if result:
                                    return result

    return None


def find_component_parent(
    components: list[dict[str, Any]],
    key: str,
) -> tuple[dict[str, Any] | None, list[dict[str, Any]], int]:
    """Find the parent container and index of a component.

    Args:
        components: List of Form.io components.
        key: The component key to find.

    Returns:
        Tuple of (parent_component, siblings_list, index).
        parent_component is None if the component is at root level.
        siblings_list is the list containing the component.
        index is the position of the component in siblings_list.

    Raises:
        ValueError: If component is not found.
        RecursionError: If nesting exceeds MAX_RECURSION_DEPTH.
    """

    def _search(
        comps: list[dict[str, Any]], parent: dict[str, Any] | None, depth: int = 0
    ) -> tuple[dict[str, Any] | None, list[dict[str, Any]], int] | None:
        if depth > MAX_RECURSION_DEPTH:
            raise RecursionError(
                f"Component tree exceeds maximum depth of {MAX_RECURSION_DEPTH}"
            )

        for i, comp in enumerate(comps):
            if not isinstance(comp, dict):
                continue

            if comp.get("key") == key:
                return (parent, comps, i)

            # Search in nested components
            nested = comp.get("components", [])
            if isinstance(nested, list) and nested:
                result = _search(nested, comp, depth + 1)
                if result:
                    return result

            # Search in columns
            columns = comp.get("columns", [])
            if isinstance(columns, list):
                for col in columns:
                    if isinstance(col, dict):
                        col_components = col.get("components", [])
                        if isinstance(col_components, list) and col_components:
                            result = _search(col_components, comp, depth + 1)
                            if result:
                                return result

            # Search in tabs
            tabs = comp.get("tabs", [])
            if isinstance(tabs, list):
                for tab in tabs:
                    if isinstance(tab, dict):
                        tab_components = tab.get("components", [])
                        if isinstance(tab_components, list) and tab_components:
                            result = _search(tab_components, comp, depth + 1)
                            if result:
                                return result

            # Search in table rows (rows[][] structure)
            rows = comp.get("rows", [])
            if isinstance(rows, list):
                for row in rows:
                    if isinstance(row, list):
                        for cell in row:
                            if isinstance(cell, dict):
                                cell_comps = cell.get("components", [])
                                if isinstance(cell_comps, list) and cell_comps:
                                    result = _search(cell_comps, comp, depth + 1)
                                    if result:
                                        return result

        return None

    result = _search(components, None)
    if result is None:
        raise ValueError(f"Component with key '{key}' not found")
    return result


def get_all_component_keys(
    components: list[dict[str, Any]],
    include_duplicates: bool = False,
) -> set[str] | list[str]:
    """Get all component keys in the tree.

    Args:
        components: List of Form.io components.
        include_duplicates: If True, returns list with duplicates preserved.
            If False (default), returns set of unique keys.

    Returns:
        Set of unique keys, or list with duplicates if include_duplicates=True.

    Raises:
        RecursionError: If nesting exceeds MAX_RECURSION_DEPTH.
    """
    # Handle non-list components (BPA API may return int or other types)
    if not isinstance(components, list):
        return [] if include_duplicates else set()

    keys: list[str] = []

    def _collect(comps: list[dict[str, Any]], depth: int = 0) -> None:
        if depth > MAX_RECURSION_DEPTH:
            raise RecursionError(
                f"Component tree exceeds maximum depth of {MAX_RECURSION_DEPTH}"
            )

        # Skip if comps is not a list (defensive for recursive calls)
        if not isinstance(comps, list):
            return

        for comp in comps:
            if not isinstance(comp, dict):
                continue

            key = comp.get("key")
            if key:
                keys.append(key)

            # Collect from nested components
            nested = comp.get("components", [])
            if nested and isinstance(nested, list):
                _collect(nested, depth + 1)

            # Collect from columns
            columns = comp.get("columns", [])
            if isinstance(columns, list):
                for col in columns:
                    if isinstance(col, dict):
                        col_components = col.get("components", [])
                        if col_components and isinstance(col_components, list):
                            _collect(col_components, depth + 1)

            # Collect from tabs
            tabs = comp.get("tabs", [])
            if isinstance(tabs, list):
                for tab in tabs:
                    if isinstance(tab, dict):
                        tab_components = tab.get("components", [])
                        if tab_components and isinstance(tab_components, list):
                            _collect(tab_components, depth + 1)

            # Collect from table rows (rows[][] structure)
            rows = comp.get("rows", [])
            if isinstance(rows, list):
                for row in rows:
                    if isinstance(row, list):
                        for cell in row:
                            if isinstance(cell, dict):
                                cell_comps = cell.get("components", [])
                                if cell_comps and isinstance(cell_comps, list):
                                    _collect(cell_comps, depth + 1)

    _collect(components)

    if include_duplicates:
        return keys
    return set(keys)


def validate_component_key_unique(
    components: list[dict[str, Any]],
    key: str,
) -> bool:
    """Check if a key is unique within the component tree.

    Args:
        components: Existing Form.io components.
        key: Key to check for uniqueness.

    Returns:
        True if key is unique, False if it already exists.
    """
    existing = get_all_component_keys(components)
    return key not in existing


def insert_component(
    components: list[dict[str, Any]],
    component: dict[str, Any],
    parent_key: str | None = None,
    position: int | None = None,
    column_index: int | str | None = None,
    row_index: int | str | None = None,
    cell_index: int | str | None = None,
) -> list[dict[str, Any]]:
    """Insert a component into the tree.

    Creates a deep copy of the components and inserts the new component.
    Does not mutate the original list.

    Args:
        components: Original Form.io components list.
        component: Component to insert.
        parent_key: Key of parent container to insert into.
                   If None, inserts at root level.
        position: Position to insert at (0-indexed).
                 If None, appends to end.
        column_index: For columns-type parents, which column to insert into.
                     Use int (0-indexed) for specific column, or "all" to
                     insert into every column. Required for columns component.
        row_index: For table-type parents, which row to insert into.
                  Use int (0-indexed) for specific row, or "all" to insert
                  into every row. Required for table component.
        cell_index: For table-type parents, which cell in the row to insert into.
                   Use int (0-indexed) for specific cell, or "all" to insert
                   into every cell. Required for table component.

    Returns:
        New components list with the component inserted.

    Raises:
        ValueError: If parent_key is specified but not found,
                   or if parent is not a container type,
                   or if column_index is missing for columns parent,
                   or if row_index/cell_index is missing for table parent,
                   or if column_index and row_index/cell_index are both specified.
    """
    # NOTE: Deep copy ensures immutability - input components are never mutated.
    # For large forms (500+ components), this may impact performance (~50-100ms).
    # If performance becomes critical, consider structural sharing optimization.
    result = copy.deepcopy(components)
    new_comp = copy.deepcopy(component)

    # Add BPA defaults if not present
    for key, value in BPA_COMPONENT_DEFAULTS.items():
        if key not in new_comp:
            new_comp[key] = copy.deepcopy(value)

    # Initialize empty components array for container types (matches backend behavior)
    if new_comp.get("type") in CONTAINER_TYPES and "components" not in new_comp:
        new_comp["components"] = []

    # Validate mutual exclusivity of column_index and row_index/cell_index
    if column_index is not None and (row_index is not None or cell_index is not None):
        raise ValueError(
            "Cannot specify both column_index and row_index/cell_index. "
            "Use column_index for columns components, "
            "or row_index/cell_index for table components."
        )

    if parent_key is None:
        # Insert at root level
        if position is None:
            result.append(new_comp)
        else:
            result.insert(position, new_comp)
    else:
        # Find parent and insert
        parent_result = find_component(result, parent_key)
        if parent_result is None:
            raise ValueError(f"Parent component '{parent_key}' not found")

        parent_comp = parent_result[0]
        parent_type = parent_comp.get("type", "")

        # Allow containers by type OR by structure (tab panes have no type
        # but contain a components[] array)
        has_components_array = isinstance(parent_comp.get("components"), list)
        if parent_type not in CONTAINER_TYPES and not has_components_array:
            raise ValueError(
                f"Component '{parent_key}' (type: {parent_type}) is not a container. "
                f"Valid container types: {', '.join(sorted(CONTAINER_TYPES))}"
            )

        # Special handling for tabs type - only tab children allowed
        # BPA backend (Jackson) enforces: Tabs.components[] must contain Tab objects.
        # Inserting a non-tab component (e.g. panel) directly crashes deserialization.
        if parent_type == "tabs":
            new_type = new_comp.get("type", "")
            if new_type != "tab":
                # List existing tab keys to help the user target the right one
                tab_keys = [
                    t.get("key", t.get("label", f"tab[{i}]"))
                    for i, t in enumerate(parent_comp.get("components", []))
                ]
                tab_hint = (
                    f" Existing tabs: {', '.join(tab_keys)}."
                    if tab_keys
                    else " No tabs exist yet — add a 'tab' component first."
                )
                raise ValueError(
                    f"Cannot add '{new_type}' directly inside a 'tabs' component. "
                    f"BPA requires tabs to contain only 'tab' children. "
                    f"Use a tab's key as parent_key to add content inside a tab."
                    f"{tab_hint}"
                )

        # Special handling for columns type
        if parent_type in COLUMN_BASED_CONTAINERS:
            columns = parent_comp.get("columns", [])
            num_columns = len(columns)

            if column_index is None:
                raise ValueError(
                    f"Parent '{parent_key}' is a columns component with "
                    f"{num_columns} columns. Specify column_index "
                    f"(0-{num_columns - 1}) to target a specific column, "
                    f'or "all" to insert into every column. '
                    f"Use form_component_get('{parent_key}') to inspect structure."
                )

            # Handle "all" - insert into every column
            if column_index == "all":
                for col_idx, col in enumerate(columns):
                    if "components" not in col:
                        col["components"] = []
                    # Deep copy for each column to avoid shared references
                    col_comp = copy.deepcopy(new_comp)
                    # Append unique suffix to key to avoid duplicates
                    if "key" in col_comp:
                        col_comp["key"] = f"{col_comp['key']}_{col_idx}"
                    if position is None:
                        col["components"].append(col_comp)
                    else:
                        col["components"].insert(position, col_comp)
                return result

            # Validate numeric column_index
            if not isinstance(column_index, int):
                raise ValueError(
                    f'column_index must be an int or "all", got: {column_index!r}'
                )

            if column_index < 0 or column_index >= num_columns:
                raise ValueError(
                    f"column_index {column_index} is out of range. "
                    f"Valid range: 0-{num_columns - 1} "
                    f"({num_columns} columns available)."
                )

            # Insert into the specified column's components array
            target_column = columns[column_index]
            if "components" not in target_column:
                target_column["components"] = []

            if position is None:
                target_column["components"].append(new_comp)
            else:
                target_column["components"].insert(position, new_comp)

            return result

        # Special handling for table type
        if parent_type in TABLE_BASED_CONTAINERS:
            rows = parent_comp.get("rows", [])
            num_rows = len(rows)

            if row_index is None or cell_index is None:
                raise ValueError(
                    f"Parent '{parent_key}' is a table component with "
                    f"{num_rows} rows. Specify both row_index and cell_index "
                    f"to target a specific cell. Use row_index=0..{num_rows - 1} "
                    f'and cell_index=0..<num_cells>, or "all" for either. '
                    f"Use form_component_get('{parent_key}') to inspect structure."
                )

            # Determine which rows to target
            if row_index == "all":
                target_row_indices = list(range(num_rows))
            else:
                if not isinstance(row_index, int):
                    raise ValueError(
                        f'row_index must be an int or "all", got: {row_index!r}'
                    )
                if row_index < 0 or row_index >= num_rows:
                    raise ValueError(
                        f"row_index {row_index} is out of range. "
                        f"Valid range: 0-{num_rows - 1} "
                        f"({num_rows} rows available)."
                    )
                target_row_indices = [row_index]

            # Process each target row
            for r_idx in target_row_indices:
                row = rows[r_idx]
                if not isinstance(row, list):
                    continue

                num_cells = len(row)

                # Determine which cells to target
                if cell_index == "all":
                    target_cell_indices = list(range(num_cells))
                else:
                    if not isinstance(cell_index, int):
                        raise ValueError(
                            f'cell_index must be an int or "all", got: {cell_index!r}'
                        )
                    if cell_index < 0 or cell_index >= num_cells:
                        raise ValueError(
                            f"cell_index {cell_index} is out of range for row {r_idx}. "
                            f"Valid range: 0-{num_cells - 1} "
                            f"({num_cells} cells in row)."
                        )
                    target_cell_indices = [cell_index]

                # Process each target cell
                for c_idx in target_cell_indices:
                    cell = row[c_idx]
                    if not isinstance(cell, dict):
                        continue

                    if "components" not in cell:
                        cell["components"] = []

                    # Deep copy for each cell to avoid shared references
                    cell_comp = copy.deepcopy(new_comp)

                    # Append suffix to avoid key duplicates across cells
                    # Both "all": key_0_0, key_0_1; one "all": key_0, key_1
                    if "key" in cell_comp:
                        if row_index == "all" and cell_index == "all":
                            cell_comp["key"] = f"{cell_comp['key']}_{r_idx}_{c_idx}"
                        elif row_index == "all":
                            cell_comp["key"] = f"{cell_comp['key']}_{r_idx}"
                        elif cell_index == "all":
                            cell_comp["key"] = f"{cell_comp['key']}_{c_idx}"
                        # If neither is "all", keep original key

                    if position is None:
                        cell["components"].append(cell_comp)
                    else:
                        cell["components"].insert(position, cell_comp)

            return result

        # Standard container handling (panel, tabs, etc.)
        # Ensure components list exists
        if "components" not in parent_comp:
            parent_comp["components"] = []

        if position is None:
            parent_comp["components"].append(new_comp)
        else:
            parent_comp["components"].insert(position, new_comp)

    return result


def remove_component(
    components: list[dict[str, Any]],
    key: str,
) -> tuple[list[dict[str, Any]], dict[str, Any]]:
    """Remove a component from the tree.

    Creates a deep copy and removes the component.

    Args:
        components: Original Form.io components list.
        key: Key of component to remove.

    Returns:
        Tuple of (new_components, removed_component).

    Raises:
        ValueError: If component is not found.
    """
    # Deep copy ensures immutability (see insert_component for performance notes)
    result = copy.deepcopy(components)

    # Find and remove
    parent, siblings, index = find_component_parent(result, key)
    removed = siblings.pop(index)

    return (result, removed)


def update_component(
    components: list[dict[str, Any]],
    key: str,
    updates: dict[str, Any],
) -> tuple[list[dict[str, Any]], dict[str, Any]]:
    """Update properties of a component.

    Merges updates into the existing component. Deep merges nested dicts.

    Args:
        components: Original Form.io components list.
        key: Key of component to update.
        updates: Properties to update/add.

    Returns:
        Tuple of (new_components, previous_state).
        previous_state contains the component's state before update.

    Raises:
        ValueError: If component is not found.
    """
    # Deep copy ensures immutability (see insert_component for performance notes)
    result = copy.deepcopy(components)

    found = find_component(result, key)
    if found is None:
        raise ValueError(f"Component with key '{key}' not found")

    comp = found[0]
    previous_state = copy.deepcopy(comp)

    # Deep merge updates
    _deep_merge(comp, updates)

    return (result, previous_state)


def move_component(
    components: list[dict[str, Any]],
    key: str,
    new_parent_key: str | None = None,
    new_position: int | None = None,
    column_index: int | str | None = None,
    row_index: int | str | None = None,
    cell_index: int | str | None = None,
) -> tuple[list[dict[str, Any]], dict[str, str | int | None]]:
    """Move a component to a new location in the tree.

    Args:
        components: Original Form.io components list.
        key: Key of component to move.
        new_parent_key: New parent container key, or None for root.
        new_position: Position in new parent, or None for end.
        column_index: For columns-type parents, which column to move into.
                     Use int (0-indexed) for specific column, or "all" to
                     move into every column (creates copies). Required for columns.
        row_index: For table-type parents, which row to move into.
                  Use int (0-indexed) for specific row, or "all" to move
                  into every row (creates copies). Required for tables.
        cell_index: For table-type parents, which cell in the row to move into.
                   Use int (0-indexed) for specific cell, or "all" to move
                   into every cell (creates copies). Required for tables.

    Returns:
        Tuple of (new_components, move_info).
        move_info contains old_parent, old_position, new_parent, new_position,
        column_index, row_index, and cell_index (if applicable).

    Raises:
        ValueError: If component or new parent not found,
                   or if column_index is missing for columns parent,
                   or if row_index/cell_index is missing for table parent.
    """
    # First remove the component
    result, removed = remove_component(components, key)

    # Track old position
    old_parent, _, old_index = find_component_parent(components, key)
    old_parent_key = old_parent.get("key") if old_parent else None

    # Insert at new location (all indices passed for columns/table support)
    result = insert_component(
        result,
        removed,
        new_parent_key,
        new_position,
        column_index,
        row_index,
        cell_index,
    )

    # Determine actual new position
    actual_position: int | None = None
    if new_parent_key:
        new_parent = find_component(result, new_parent_key)
        if new_parent:
            parent_comp = new_parent[0]
            parent_type = parent_comp.get("type", "")
            # For columns, get position from specific column
            if parent_type in COLUMN_BASED_CONTAINERS and column_index is not None:
                columns = parent_comp.get("columns", [])
                # For "all", use first column for position calculation
                col_idx = 0 if column_index == "all" else column_index
                if isinstance(col_idx, int) and 0 <= col_idx < len(columns):
                    col_components = columns[col_idx].get("components", [])
                    actual_position = (
                        new_position
                        if new_position is not None
                        else len(col_components) - 1
                    )
            # For tables, get position from specific cell
            elif parent_type in TABLE_BASED_CONTAINERS and row_index is not None:
                rows = parent_comp.get("rows", [])
                # For "all", use first row/cell for position calculation
                r_idx = 0 if row_index == "all" else row_index
                c_idx = 0 if cell_index == "all" else cell_index
                if isinstance(r_idx, int) and isinstance(c_idx, int):
                    if 0 <= r_idx < len(rows) and isinstance(rows[r_idx], list):
                        row = rows[r_idx]
                        if 0 <= c_idx < len(row) and isinstance(row[c_idx], dict):
                            cell_components = row[c_idx].get("components", [])
                            actual_position = (
                                new_position
                                if new_position is not None
                                else len(cell_components) - 1
                            )
            else:
                new_siblings = parent_comp.get("components", [])
                actual_position = (
                    new_position if new_position is not None else len(new_siblings) - 1
                )
        else:
            actual_position = new_position
    else:
        actual_position = new_position if new_position is not None else len(result) - 1

    move_info: dict[str, str | int | None] = {
        "old_parent": old_parent_key,
        "old_position": old_index,
        "new_parent": new_parent_key,
        "new_position": actual_position,
        "column_index": column_index,
        "row_index": row_index,
        "cell_index": cell_index,
    }

    return (result, move_info)


def _deep_merge(target: dict[str, Any], source: dict[str, Any]) -> None:
    """Deep merge source dict into target dict (mutates target).

    Args:
        target: Dict to merge into.
        source: Dict to merge from.
    """
    for key, value in source.items():
        if key in target and isinstance(target[key], dict) and isinstance(value, dict):
            _deep_merge(target[key], value)
        else:
            target[key] = copy.deepcopy(value)


# =============================================================================
# Component Builders - Convenience functions for creating common components
# =============================================================================


def build_textfield(
    key: str,
    label: str,
    *,
    required: bool = False,
    placeholder: str = "",
    tooltip: str = "",
    description: str = "",
    size: str = "md",
    hidden: bool = False,
    disabled: bool = False,
    default_value: str = "",
) -> dict[str, Any]:
    """Build a textfield component with BPA-specific properties.

    Args:
        key: Unique component key (e.g., "applicantFullName").
        label: Display label.
        required: Whether field is required.
        placeholder: Placeholder text.
        tooltip: Tooltip text.
        description: Field description.
        size: Field size ("sm", "md", "lg").
        hidden: Whether field is hidden.
        disabled: Whether field is disabled.
        default_value: Default value.

    Returns:
        Complete textfield component definition.
    """
    component: dict[str, Any] = {
        "type": "textfield",
        "key": key,
        "label": label,
        "size": size,
        "validate": {"required": required},
        "input": True,
        "tableView": True,
        **BPA_COMPONENT_DEFAULTS,
    }

    if placeholder:
        component["placeholder"] = placeholder
    if tooltip:
        component["tooltip"] = tooltip
    if description:
        component["description"] = description
    if hidden:
        component["hidden"] = True
    if disabled:
        component["disabled"] = True
    if default_value:
        component["defaultValue"] = default_value

    return component


def build_number(
    key: str,
    label: str,
    *,
    required: bool = False,
    min_value: float | None = None,
    max_value: float | None = None,
    decimal_limit: int | None = None,
    default_value: float | None = None,
) -> dict[str, Any]:
    """Build a number component.

    Args:
        key: Unique component key.
        label: Display label.
        required: Whether field is required.
        min_value: Minimum allowed value.
        max_value: Maximum allowed value.
        decimal_limit: Maximum decimal places.
        default_value: Default value.

    Returns:
        Complete number component definition.
    """
    component: dict[str, Any] = {
        "type": "number",
        "key": key,
        "label": label,
        "validate": {"required": required},
        "input": True,
        "tableView": True,
        **BPA_COMPONENT_DEFAULTS,
    }

    if min_value is not None:
        component["validate"]["min"] = min_value
    if max_value is not None:
        component["validate"]["max"] = max_value
    if decimal_limit is not None:
        component["decimalLimit"] = decimal_limit
    if default_value is not None:
        component["defaultValue"] = default_value

    return component


def build_select(
    key: str,
    label: str,
    *,
    required: bool = False,
    data_source: str = "values",
    values: list[dict[str, str]] | None = None,
    catalog: str | None = None,
    multiple: bool = False,
    placeholder: str = "",
) -> dict[str, Any]:
    """Build a select/dropdown component.

    Args:
        key: Unique component key.
        label: Display label.
        required: Whether field is required.
        data_source: Data source type ("values", "Catalogue", "url").
        values: List of {"label": ..., "value": ...} options.
        catalog: Catalog name (when data_source is "Catalogue").
        multiple: Allow multiple selections.
        placeholder: Placeholder text.

    Returns:
        Complete select component definition.
    """
    component: dict[str, Any] = {
        "type": "select",
        "key": key,
        "label": label,
        "validate": {"required": required},
        "input": True,
        "tableView": True,
        "data": {"dataSrc": data_source},
        **BPA_COMPONENT_DEFAULTS,
    }

    if data_source == "values" and values:
        component["data"]["values"] = values
    elif data_source == "Catalogue" and catalog:
        component["data"]["catalog"] = catalog
        component["data"]["systemSource"] = "BPA"

    if multiple:
        component["multiple"] = True
    if placeholder:
        component["placeholder"] = placeholder

    return component


def build_checkbox(
    key: str,
    label: str,
    *,
    required: bool = False,
    default_value: bool = False,
) -> dict[str, Any]:
    """Build a checkbox component.

    Args:
        key: Unique component key.
        label: Display label.
        required: Whether field is required.
        default_value: Default checked state.

    Returns:
        Complete checkbox component definition.
    """
    return {
        "type": "checkbox",
        "key": key,
        "label": label,
        "validate": {"required": required},
        "defaultValue": default_value,
        "input": True,
        "tableView": True,
        **BPA_COMPONENT_DEFAULTS,
    }


def build_textarea(
    key: str,
    label: str,
    *,
    required: bool = False,
    rows: int = 3,
    placeholder: str = "",
) -> dict[str, Any]:
    """Build a textarea component.

    Args:
        key: Unique component key.
        label: Display label.
        required: Whether field is required.
        rows: Number of visible rows.
        placeholder: Placeholder text.

    Returns:
        Complete textarea component definition.
    """
    component: dict[str, Any] = {
        "type": "textarea",
        "key": key,
        "label": label,
        "rows": rows,
        "validate": {"required": required},
        "input": True,
        "tableView": True,
        **BPA_COMPONENT_DEFAULTS,
    }

    if placeholder:
        component["placeholder"] = placeholder

    return component


def build_panel(
    key: str,
    title: str,
    *,
    components: list[dict[str, Any]] | None = None,
    collapsible: bool = False,
    collapsed: bool = False,
) -> dict[str, Any]:
    """Build a panel container component.

    Args:
        key: Unique component key.
        title: Panel title.
        components: Nested components.
        collapsible: Whether panel can be collapsed.
        collapsed: Initial collapsed state.

    Returns:
        Complete panel component definition.
    """
    panel: dict[str, Any] = {
        "type": "panel",
        "key": key,
        "title": title,
        "components": components or [],
        "input": False,
        "tableView": False,
        **BPA_COMPONENT_DEFAULTS,
    }

    if collapsible:
        panel["collapsible"] = True
        panel["collapsed"] = collapsed

    return panel


def build_columns(
    key: str,
    *,
    column_count: int = 2,
    components_per_column: list[list[dict[str, Any]]] | None = None,
) -> dict[str, Any]:
    """Build a columns layout component.

    Args:
        key: Unique component key.
        column_count: Number of columns (default 2).
        components_per_column: List of component lists for each column.

    Returns:
        Complete columns component definition.
    """
    columns = []
    width = 12 // column_count  # Bootstrap grid (12 columns)

    for i in range(column_count):
        col_components = (
            components_per_column[i]
            if components_per_column and i < len(components_per_column)
            else []
        )
        columns.append(
            {
                "components": col_components,
                "width": width,
                "offset": 0,
                "push": 0,
                "pull": 0,
                "size": "md",
            }
        )

    return {
        "type": "columns",
        "key": key,
        "columns": columns,
        "input": False,
        "tableView": False,
        **BPA_COMPONENT_DEFAULTS,
    }


# =============================================================================
# Validation Functions
# =============================================================================


def validate_component(component: dict[str, Any]) -> list[str]:
    """Validate a component structure.

    Args:
        component: Component to validate.

    Returns:
        List of validation error messages (empty if valid).
    """
    errors: list[str] = []

    # Required fields
    if not component.get("key"):
        errors.append("Component must have a 'key' property")
    elif not _is_valid_key(component["key"]):
        errors.append(
            f"Invalid key format '{component['key']}'. "
            "Keys must be alphanumeric with underscores, starting with a letter."
        )

    if not component.get("type"):
        errors.append("Component must have a 'type' property")

    # Type-specific validation
    comp_type = component.get("type", "")

    if comp_type in ("textfield", "textarea", "number", "select", "checkbox"):
        if not component.get("label"):
            errors.append(
                f"Component type '{comp_type}' should have a 'label' property"
            )

    if comp_type == "select":
        data = component.get("data", {})
        data_src = data.get("dataSrc", "values")
        if data_src == "values" and not data.get("values"):
            # Values source but no values - just a warning
            pass
        elif data_src == "Catalogue" and not data.get("catalog"):
            errors.append(
                "Select with Catalogue source must have 'data.catalog' property"
            )

    return errors


def _is_valid_key(key: str) -> bool:
    """Check if a key follows valid naming conventions.

    Args:
        key: Component key to validate.

    Returns:
        True if key is valid.
    """
    # Must start with letter, contain only alphanumeric and underscores
    return bool(re.match(r"^[a-zA-Z][a-zA-Z0-9_]*$", key))


def validate_form_schema(schema: dict[str, Any]) -> list[str]:
    """Validate a complete form schema.

    Args:
        schema: Form schema with components.

    Returns:
        List of validation error messages.
    """
    errors: list[str] = []

    if "components" not in schema:
        errors.append("Form schema must have 'components' array")
        return errors

    components = schema["components"]
    if not isinstance(components, list):
        errors.append("'components' must be an array")
        return errors

    # Check for duplicate keys using list with duplicates preserved
    all_keys = get_all_component_keys(components, include_duplicates=True)
    seen: set[str] = set()
    for key in all_keys:
        if key in seen:
            errors.append(f"Duplicate component key: '{key}'")
        else:
            seen.add(key)

    # Validate each component
    for i, comp in enumerate(components):
        comp_errors = validate_component(comp)
        for error in comp_errors:
            errors.append(f"Component [{i}]: {error}")

    return errors



"""MCP tools for BPA field operations.

This module provides tools for listing and retrieving BPA form fields.
Fields are accessed through service endpoints (service-centric API design).
"""

from __future__ import annotations

from typing import Any

from fastmcp.exceptions import ToolError

from mcp_eregistrations_bpa.bpa_client import BPAClient
from mcp_eregistrations_bpa.bpa_client.errors import (
    BPAClientError,
    BPANotFoundError,
    translate_error,
)
from mcp_eregistrations_bpa.tools.large_response import large_response_handler

__all__ = ["field_list", "field_get", "register_field_tools"]


@large_response_handler(
    threshold_bytes=50 * 1024,  # 50KB threshold for list tools
    navigation={
        "list_all": "jq '.fields'",
        "find_by_type": "jq '.fields[] | select(.type == \"textfield\")'",
        "find_by_key": "jq '.fields[] | select(.key | contains(\"search\"))'",
        "required_only": "jq '.fields[] | select(.required == true)'",
    },
)
async def field_list(
    service_id: str | int,
    limit: int = 50,
    offset: int = 0,
    type_filter: str | None = None,
    required_only: bool = False,
    instance: str | None = None,
) -> dict[str, Any]:
    """List fields for a service with pagination and filtering.

    Large responses (>50KB) are saved to file with navigation hints.

    Args:
        service_id: The service ID to list fields for.
        limit: Maximum number of fields to return (default: 50).
        offset: Number of fields to skip (default: 0).
        type_filter: Filter by field type (e.g., "select", "textfield").
        required_only: Filter to only required fields (default: False).
        instance: Instance profile name (from instance_list).

    Returns:
        dict with fields, total, has_more, service_id.
    """
    # Normalize pagination parameters
    if limit <= 0:
        limit = 50
    if offset < 0:
        offset = 0

    try:
        async with BPAClient.for_instance(instance) as client:
            fields_data = await client.get_list(
                "/service/{service_id}/fields",
                path_params={"service_id": service_id},
                resource_type="field",
            )
    except BPAClientError as e:
        raise translate_error(e, resource_type="field")

    # Transform to consistent output format with snake_case keys
    all_fields = []
    for field in fields_data:
        field_obj: dict[str, Any] = {
            "key": field.get("key"),
            "name": field.get("name"),
            "type": field.get("type"),
            "required": field.get("required", False),
            "label": field.get("label"),
        }
        # Only include component_key if it has a value (remove null noise)
        if field.get("componentKey"):
            field_obj["component_key"] = field.get("componentKey")
        all_fields.append(field_obj)

    # Apply filters before sorting/pagination
    if type_filter is not None:
        all_fields = [f for f in all_fields if f.get("type") == type_filter]

    if required_only:
        all_fields = [f for f in all_fields if f.get("required") is True]

    # Sort by key for consistent pagination ordering
    all_fields.sort(key=lambda f: f.get("key") or "")

    # Calculate total before pagination
    total = len(all_fields)

    # Apply pagination
    paginated_fields = all_fields[offset : offset + limit]

    # Calculate has_more
    has_more = (offset + limit) < total

    return {
        "fields": paginated_fields,
        "total": total,
        "has_more": has_more,
        "service_id": service_id,
    }


async def field_get(
    service_id: str | int, field_key: str, instance: str | None = None
) -> dict[str, Any]:
    """Get details of a BPA field by service ID and field key.

    Args:
        service_id: The service containing the field.
        field_key: The field key/identifier within the service.
        instance: Instance profile name (from instance_list).

    Returns:
        dict with key, name, label, type, required, component_key, service_id.
    """
    try:
        async with BPAClient.for_instance(instance) as client:
            try:
                field_data = await client.get(
                    "/service/{service_id}/fields/{field_key}",
                    path_params={"service_id": service_id, "field_key": field_key},
                    resource_type="field",
                    resource_id=field_key,
                )
            except BPANotFoundError:
                raise ToolError(
                    f"Field '{field_key}' not found in service '{service_id}'. "
                    "Use 'field_list' with the service_id to see available fields."
                )
    except ToolError:
        raise
    except BPAClientError as e:
        raise translate_error(e, resource_type="field", resource_id=field_key)

    result = {
        "key": field_data.get("key"),
        "name": field_data.get("name"),
        "label": field_data.get("label"),
        "type": field_data.get("type"),
        "required": field_data.get("required", False),
        "service_id": service_id,
    }
    # Only include component_key if it has a value (remove null noise)
    if field_data.get("componentKey"):
        result["component_key"] = field_data.get("componentKey")
    return result


def register_field_tools(mcp: Any) -> None:
    """Register field tools with the MCP server.

    Args:
        mcp: The FastMCP server instance.
    """
    from mcp_eregistrations_bpa.tools.annotations import READ

    mcp.tool(annotations=READ)(field_list)
    mcp.tool(annotations=READ)(field_get)



"""HTTP client for the DS Django REST Framework API."""

from __future__ import annotations

import secrets
from typing import Any

import httpx

from mcp_eregistrations_ds.ds_client.errors import (
    DSAuthenticationError,
    DSClientError,
    DSConnectionError,
    DSNotFoundError,
    DSPermissionError,
    DSServerError,
    DSTimeoutError,
)


class DSClient:
    """Async HTTP client for DS REST API.

    Usage::

        async with DSClient(base_url="https://jamaica.eregistrations.org") as c:
            files = await c.get_paginated("/backend/admin/files")
    """

    def __init__(
        self,
        base_url: str,
        token: str | None = None,
        timeout: float = 30.0,
        verify_ssl: bool = True,
    ):
        self._base_url = base_url.rstrip("/")
        self._token = token
        self._timeout = timeout
        self._verify_ssl = verify_ssl
        self._client: httpx.AsyncClient | None = None
        self._instance: str | None = None  # Set by for_instance()

    async def __aenter__(self) -> DSClient:
        # Try shared keyring if no in-memory token
        if not self._token and self._instance:
            from mcp_eregistrations_common.auth import get_or_refresh_token

            self._token = await get_or_refresh_token(self._instance)

        headers: dict[str, str] = {}
        if self._token:
            headers["Authorization"] = f"Bearer {self._token}"

        # Generate self-signed CSRF token pair.
        # DRF SessionAuthentication.enforce_csrf() compares the csrftoken
        # cookie against the X-CSRFToken header — a matching pair satisfies it.
        # Django also requires Referer header for HTTPS CSRF validation.
        csrf_token = secrets.token_hex(32)
        headers["X-CSRFToken"] = csrf_token
        headers["Referer"] = self._base_url + "/"

        self._client = httpx.AsyncClient(
            base_url=self._base_url,
            headers=headers,
            timeout=self._timeout,
            cookies={"csrftoken": csrf_token},
            verify=self._verify_ssl,
        )
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self._client:
            await self._client.aclose()
            self._client = None

    async def _refresh_and_update_header(self) -> bool:
        """Attempt token refresh and update client headers."""
        if not self._instance:
            return False
        from mcp_eregistrations_common.auth import get_or_refresh_token

        new_token = await get_or_refresh_token(self._instance)
        if new_token and new_token != self._token:
            self._token = new_token
            assert self._client is not None
            self._client.headers["Authorization"] = f"Bearer {new_token}"
            return True
        return False

    async def _request_with_retry(
        self, method: str, path: str, **kwargs: Any
    ) -> httpx.Response:
        """Execute request with one retry on 401 (token refresh)."""
        assert self._client is not None
        response = await self._client.request(method, path, **kwargs)
        if response.status_code == 401 and self._instance:
            refreshed = await self._refresh_and_update_header()
            if refreshed:
                response = await self._client.request(method, path, **kwargs)
        return response

    def _raise_for_status(self, response: httpx.Response) -> None:
        """Translate HTTP status codes to DS errors."""
        if response.is_success:
            return
        status = response.status_code
        detail = ""
        try:
            detail = response.json().get("detail", response.text)
        except Exception:
            detail = response.text

        if status == 401:
            raise DSAuthenticationError(detail, status_code=status)
        if status == 403:
            raise DSPermissionError(detail, status_code=status)
        if status == 404:
            raise DSNotFoundError(detail, status_code=status)
        if status >= 500:
            raise DSServerError(detail, status_code=status)
        # Generic 4xx client error
        raise DSClientError(f"HTTP {status}: {detail}", status_code=status)

    async def get(
        self,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> Any:
        """GET request returning JSON response."""
        assert self._client is not None, (
            "DSClient must be used as async context manager"
        )
        kwargs: dict[str, Any] = {"params": params}
        if timeout is not None:
            kwargs["timeout"] = httpx.Timeout(timeout)
        try:
            response = await self._request_with_retry("GET", path, **kwargs)
        except httpx.ConnectError as e:
            raise DSConnectionError(str(e)) from e
        except httpx.TimeoutException as e:
            raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        return response.json()

    async def post(
        self,
        path: str,
        *,
        json: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """POST request with JSON body."""
        assert self._client is not None, (
            "DSClient must be used as async context manager"
        )
        try:
            response = await self._request_with_retry(
                "POST", path, json=json, params=params
            )
        except httpx.ConnectError as e:
            raise DSConnectionError(str(e)) from e
        except httpx.TimeoutException as e:
            raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        return response.json()

    async def put(
        self,
        path: str,
        *,
        json: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """PUT request with JSON body."""
        assert self._client is not None, (
            "DSClient must be used as async context manager"
        )
        try:
            response = await self._request_with_retry(
                "PUT", path, json=json, params=params
            )
        except httpx.ConnectError as e:
            raise DSConnectionError(str(e)) from e
        except httpx.TimeoutException as e:
            raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        return response.json()

    async def patch(
        self,
        path: str,
        *,
        json: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """PATCH request with JSON body."""
        assert self._client is not None, (
            "DSClient must be used as async context manager"
        )
        try:
            response = await self._request_with_retry(
                "PATCH", path, json=json, params=params
            )
        except httpx.ConnectError as e:
            raise DSConnectionError(str(e)) from e
        except httpx.TimeoutException as e:
            raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        return response.json()

    async def delete(
        self,
        path: str,
        *,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """DELETE request."""
        assert self._client is not None, (
            "DSClient must be used as async context manager"
        )
        try:
            response = await self._request_with_retry("DELETE", path, params=params)
        except httpx.ConnectError as e:
            raise DSConnectionError(str(e)) from e
        except httpx.TimeoutException as e:
            raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        if response.status_code == 204 or not response.content:
            return {}
        return response.json()

    async def get_bytes(
        self,
        url: str,
        timeout: float = 60.0,
    ) -> tuple[bytes, str]:
        """GET request returning raw bytes and content-type.

        Relative URLs use the authenticated client. Absolute URLs
        (e.g. MinIO presigned) use a plain client without auth headers.

        Args:
            url: Document URL — relative (Django-served) or absolute (MinIO presigned).
            timeout: Per-request timeout in seconds (default: 60).
        """
        if url.startswith(("http://", "https://")):
            # Absolute URL (MinIO presigned) — no auth headers
            try:
                async with httpx.AsyncClient() as ext_client:
                    response = await ext_client.get(url, timeout=timeout)
            except httpx.ConnectError as e:
                raise DSConnectionError(str(e)) from e
            except httpx.TimeoutException as e:
                raise DSTimeoutError(str(e)) from e
        else:
            # Relative URL — use authenticated client
            assert self._client is not None, (
                "DSClient must be used as async context manager"
            )
            try:
                response = await self._request_with_retry("GET", url, timeout=timeout)
            except httpx.ConnectError as e:
                raise DSConnectionError(str(e)) from e
            except httpx.TimeoutException as e:
                raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        content_type = response.headers.get("content-type", "application/octet-stream")
        return response.content, content_type

    async def upload(
        self,
        path: str,
        *,
        file_data: bytes,
        file_name: str,
        field_name: str = "file",
        extra_fields: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Multipart file upload."""
        assert self._client is not None, (
            "DSClient must be used as async context manager"
        )
        files = {field_name: (file_name, file_data)}
        data = extra_fields or {}
        try:
            response = await self._request_with_retry(
                "POST", path, files=files, data=data, params=params
            )
        except httpx.ConnectError as e:
            raise DSConnectionError(str(e)) from e
        except httpx.TimeoutException as e:
            raise DSTimeoutError(str(e)) from e
        self._raise_for_status(response)
        return response.json()

    async def get_paginated(
        self,
        path: str,
        *,
        params: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        """GET for DRF paginated endpoints.

        Normalizes: results->items, adds has_next/page.
        """
        params = params or {}
        page = params.get("page", 1)
        raw = await self.get(path, params=params, timeout=timeout)

        # DRF PageNumberPagination format
        if isinstance(raw, dict) and "results" in raw:
            items = raw["results"]
            return {
                "items": items,
                "count": raw.get("count", len(items)),
                "page": page,
                "page_size": len(items),
                "has_next": raw.get("next") is not None,
            }
        # Non-paginated list response — apply client-side pagination
        if isinstance(raw, list):
            default_page_size = 20
            total = len(raw)
            start = (page - 1) * default_page_size
            end = start + default_page_size
            items = raw[start:end]
            return {
                "items": items,
                "count": total,
                "page": page,
                "page_size": len(items),
                "has_next": end < total,
            }
        # Already a dict without "results" key
        result: dict[str, Any] = raw
        return result

    @classmethod
    def for_instance(cls, instance: str | None = None) -> DSClient:
        """Create DSClient from instance profile config.

        Looks up ds_url from common config (built-in or custom profiles).
        Requires an explicit instance name — no env var fallback.
        Token is retrieved from DS's own in-process token store
        (populated by ds_auth_login), with keyring fallback on __aenter__.
        """
        from fastmcp.exceptions import ToolError

        from mcp_eregistrations_common.config import get_instance_config

        if instance is None:
            raise ToolError(
                "No instance specified. Pass instance='<name>'. "
                "Use instance_list to see available instances."
            )

        config = get_instance_config(instance)
        ds_url = config.get("ds_url")

        if not ds_url:
            raise ToolError(
                f"No DS URL configured for instance '{instance}'. "
                f"Use instance_add to register an instance."
            )

        token = _ds_tokens.get(instance)
        client = cls(
            base_url=ds_url,
            token=token,
            verify_ssl=config.get("verify_ssl", True),
        )
        client._instance = instance
        return client


# DS-local in-process token store (populated by ds_auth_login)
_ds_tokens: dict[str, str] = {}


def store_ds_token(instance: str, token: str) -> None:
    """Store an access token for a DS instance."""
    _ds_tokens[instance] = token


def get_ds_token(instance: str) -> str | None:
    """Get stored access token for a DS instance."""
    return _ds_tokens.get(instance)



"""Shared instance configuration for all MCP eRegistrations components.

Provides:
- Built-in instance definitions from instances.yaml
- Unified profile management (~/.config/mcp-eregistrations/profiles.json)
- Instance name resolution (profiles override built-in)
- Auth key resolution (profile name lookup for authentication)
"""

from __future__ import annotations

import hashlib
import json
import logging
import re
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import yaml

logger = logging.getLogger(__name__)


def _generate_instance_slug(url: str) -> str:
    """Generate a filesystem-safe slug from a BPA instance URL.

    Creates a human-readable identifier from the hostname, plus a short hash
    for uniqueness. Format: {sanitized_hostname}-{short_hash}

    Examples:
        https://bpa.dev.els.eregistrations.org -> els-dev-a1b2c3
        https://bpa.test.cuba.eregistrations.org -> cuba-test-d4e5f6

    Args:
        url: The BPA instance URL.

    Returns:
        A filesystem-safe slug like "els-dev-a1b2c3".
    """
    parsed = urlparse(url)
    hostname = parsed.netloc or parsed.path  # Handle edge cases

    # Extract meaningful parts from hostname
    # e.g., "bpa.dev.els.eregistrations.org" ->
    # ["bpa", "dev", "els", "eregistrations", "org"]
    parts = hostname.lower().split(".")

    # Remove common prefixes/suffixes for cleaner slug
    skip_parts = {"bpa", "eregistrations", "org", "com", "www"}
    meaningful_parts = [p for p in parts if p not in skip_parts and p]

    # Build readable slug (reversed to get country/env first: "els-dev")
    if meaningful_parts:
        slug_base = "-".join(reversed(meaningful_parts[:2]))  # Max 2 parts
    else:
        slug_base = "default"

    # Add short hash for uniqueness (handles edge cases)
    url_hash = hashlib.sha256(url.encode()).hexdigest()[:6]
    slug = f"{slug_base}-{url_hash}"

    # Sanitize: only allow alphanumeric and hyphens
    slug = re.sub(r"[^a-z0-9-]", "-", slug)
    slug = re.sub(r"-+", "-", slug).strip("-")

    return slug


class ConfigurationError(Exception):
    """Raised when instance configuration is invalid or missing."""

    pass


# Patchable base dir for tests (monkeypatch this to tmp_path)
CONFIG_BASE_DIR = Path.home() / ".config"


def load_builtin_instances() -> list[dict[str, Any]]:
    """Load built-in instance definitions from common instances.yaml.

    Returns:
        List of instance dicts with name, display_name, URLs, auth config.
    """
    yaml_path = Path(__file__).parent / "instances.yaml"
    with open(yaml_path) as f:
        data = yaml.safe_load(f)
    result: list[dict[str, Any]] = data.get("instances", [])
    return result


def get_builtin_instance(name: str) -> dict[str, Any] | None:
    """Find a built-in instance by name.

    Args:
        name: Instance name (e.g., "jamaica").

    Returns:
        Instance config dict, or None if not found.
    """
    for inst in load_builtin_instances():
        if inst["name"] == name:
            return inst
    return None


# Unified profile store path
PROFILES_PATH = CONFIG_BASE_DIR / "mcp-eregistrations" / "profiles.json"

# Migration state — set to True after migration runs (or in tests)
_profiles_migrated = False


def _legacy_profiles_path(component: str) -> Path:
    """Get legacy profiles.json path for a component.

    Args:
        component: Component name ("bpa", "ds", "gdb").

    Returns:
        Path like ~/.config/mcp-eregistrations-ds/profiles.json
    """
    return CONFIG_BASE_DIR / f"mcp-eregistrations-{component}" / "profiles.json"


def _migrate_profiles() -> None:
    """Migrate legacy per-component profiles into the unified store.

    Reads BPA legacy profiles (wrapped format: {"profiles": {"name": {...}}})
    and DS legacy profiles (flat format: {"name": {...}}), merges them,
    and writes the unified store. Old files are left in place.

    All errors are logged as warnings and never block execution.
    """
    unified_path = CONFIG_BASE_DIR / "mcp-eregistrations" / "profiles.json"
    if unified_path.exists():
        return

    merged: dict[str, dict[str, Any]] = {}

    # Read BPA legacy profiles
    bpa_path = _legacy_profiles_path("bpa")
    if bpa_path.exists():
        try:
            raw = json.loads(bpa_path.read_text())
            # BPA format: {"profiles": {"name": {...}}} — note the wrapper
            bpa_profiles: dict[str, dict[str, Any]] = {}
            if isinstance(raw, dict):
                if "profiles" in raw and isinstance(raw["profiles"], dict):
                    bpa_profiles = raw["profiles"]
                else:
                    # Flat format (no wrapper)
                    bpa_profiles = raw

            for name, profile in bpa_profiles.items():
                cleaned = dict(profile)
                # Rename bpa_instance_url -> bpa_url
                if "bpa_instance_url" in cleaned:
                    cleaned["bpa_url"] = cleaned.pop("bpa_instance_url")
                # Remove keyring sentinels
                if cleaned.get("cas_client_secret") == "__keyring__":
                    del cleaned["cas_client_secret"]
                merged[name] = cleaned
        except Exception:
            logger.warning("Failed to read BPA legacy profiles from %s", bpa_path)

    # Read DS legacy profiles
    ds_path = _legacy_profiles_path("ds")
    if ds_path.exists():
        try:
            ds_profiles: dict[str, dict[str, Any]] = json.loads(ds_path.read_text())
            if isinstance(ds_profiles, dict):
                for name, profile in ds_profiles.items():
                    if name in merged:
                        # Fill in DS fields absent from BPA profile
                        for key, value in profile.items():
                            if key not in merged[name]:
                                merged[name][key] = value
                    else:
                        # DS-only profile
                        merged[name] = dict(profile)
        except Exception:
            logger.warning("Failed to read DS legacy profiles from %s", ds_path)

    if merged:
        try:
            unified_path.parent.mkdir(parents=True, exist_ok=True)
            tmp = Path(str(unified_path) + ".tmp")
            tmp.write_text(json.dumps(merged, indent=2))
            tmp.replace(unified_path)
        except Exception:
            logger.warning("Failed to write unified profiles to %s", unified_path)


def _ensure_profiles_migrated() -> None:
    """Run migration once per process if not already done."""
    global _profiles_migrated  # noqa: PLW0603
    if not _profiles_migrated:
        _migrate_profiles()
        _profiles_migrated = True


def load_profiles() -> dict[str, dict[str, Any]]:
    """Load saved instance profiles from the unified store.

    Returns:
        Dict of profile_name -> config dict.
    """
    _ensure_profiles_migrated()
    path = CONFIG_BASE_DIR / "mcp-eregistrations" / "profiles.json"
    if not path.exists():
        return {}
    try:
        result: dict[str, dict[str, Any]] = json.loads(path.read_text())
        return result
    except (json.JSONDecodeError, OSError):
        return {}


def save_profiles(profiles: dict[str, dict[str, Any]]) -> None:
    """Save instance profiles to the unified store.

    Args:
        profiles: Dict of profile_name -> config dict.
    """
    _ensure_profiles_migrated()
    path = CONFIG_BASE_DIR / "mcp-eregistrations" / "profiles.json"
    path.parent.mkdir(parents=True, exist_ok=True)
    tmp = Path(str(path) + ".tmp")
    tmp.write_text(json.dumps(profiles, indent=2))
    tmp.replace(path)


def get_instance_config(
    instance: str | None,
) -> dict[str, Any]:
    """Resolve instance name to config dict.

    Lookup order:
    1. Custom profiles (unified store)
    2. Built-in instances.yaml

    Args:
        instance: Instance name (e.g., "jamaica"). None raises ToolError.

    Returns:
        Config dict with URLs, auth config, etc.

    Raises:
        ToolError: If instance is None or not found.
    """
    from fastmcp.exceptions import ToolError

    if instance is None:
        raise ToolError(
            "No instance specified. Pass instance='<name>'. "
            "Use instance_list to see available instances."
        )

    # Check custom profiles first
    profiles = load_profiles()
    if instance in profiles:
        return profiles[instance]

    # Fall back to built-in instances
    builtin = get_builtin_instance(instance)
    if builtin is not None:
        return builtin

    available = [i["name"] for i in load_builtin_instances()]
    raise ToolError(
        f"Unknown instance '{instance}'. "
        f"Available: {', '.join(available)}. "
        f"Or run instance_add to register a custom instance."
    )


def resolve_auth_key(instance: str | None) -> str:
    """Resolve the universal auth key (profile name) for an instance.

    1. Explicit name -> return as-is
    2. None + exactly one profile -> auto-select
    3. None + multiple -> raise with list

    Args:
        instance: Instance name, or None for auto-select.

    Returns:
        Profile name to use as auth key.

    Raises:
        ConfigurationError: If auto-select fails (multiple or no instances).
    """
    if instance:
        return instance

    # Auto-select from unified profiles
    profiles = load_profiles()
    names = list(profiles.keys())
    if len(names) == 1:
        return names[0]
    if names:
        raise ConfigurationError(
            f"Multiple instances configured: {', '.join(names)}. "
            "Specify instance= parameter."
        )

    # No profiles — check built-in instances
    instances = load_builtin_instances()
    if len(instances) == 1:
        name: str = instances[0]["name"]
        return name

    raise ConfigurationError("No instance configured. Use instance_add first.")



"""MCP tool for generating normalized AI-readable service maps.

Composes a single deterministic artifact from BPA export data, reusing
existing normalization from service_insights and _form_shared.

NOTE: This module imports private functions from service_insights.py:
  _build_branch_map, _build_form_lookup, _unwrap_export
If those functions are refactored, tests in test_service_map.py will catch breakage.
"""

from __future__ import annotations

import json
import logging
from collections import defaultdict
from typing import Any

from fastmcp.exceptions import ToolError

from mcp_eregistrations_bpa.bpa_client import BPAClient
from mcp_eregistrations_bpa.bpa_client.errors import (
    BPAClientError,
    BPANotFoundError,
    translate_error,
)
from mcp_eregistrations_bpa.tools._form_shared import (
    parse_form_schema,
    simplify_components,
)
from mcp_eregistrations_bpa.tools.formio_helpers import get_all_component_keys
from mcp_eregistrations_bpa.tools.large_response import large_response_handler
from mcp_eregistrations_bpa.tools.service_insights import (
    _STATUS_TYPE_TO_EDGE,
    _build_branch_map,
    _build_form_lookup,
    _unwrap_export,
)

logger = logging.getLogger(__name__)

__all__ = [
    "service_ai_map",
    "register_service_map_tools",
]

# Export options: union of workflow_graph + service_branches needs
_AI_MAP_EXPORT_OPTIONS = {
    "serviceSelected": True,
    "rolesSelected": True,
    "determinantsSelected": True,
    "guideFormSelected": True,
    "applicantFormSelected": True,
    "sendFileFormSelected": True,
    "paymentFormSelected": True,
    "registrationsSelected": True,
    "catalogsSelected": False,
    "botsSelected": False,
    "printDocumentsSelected": False,
    "costsSelected": False,
    "requirementsSelected": False,
    "resultsSelected": False,
    "activityConditionsSelected": False,
    "registrationLawsSelected": False,
    "serviceLocationsSelected": False,
    "serviceTutorialsSelected": False,
    "serviceTranslationsSelected": False,
    "copyService": False,
}


_AGENT_TYPE_MAP = {
    "BotRole": "bot",
    "UserRole": "human",
}


def _build_intended_flow(
    roles: list[dict[str, Any]],
    warnings: list[str],
) -> dict[str, Any]:
    """Build the intended_flow section from export roles data.

    Extracts typed edges from role statuses, then detects parallel groups
    (fan-out), convergence roles (fan-in), and correction paths.
    """
    if not roles:
        warnings.append("Service has no roles configured")
        return {
            "roles": [],
            "edges": [],
            "start_roles": [],
            "parallel_groups": [],
            "convergence_roles": [],
            "correction_paths": [],
        }

    # Build edges first so we can compute used_in_flow from the graph
    edges: list[dict[str, Any]] = []
    forward_graph: dict[str, list[str]] = defaultdict(list)
    inbound_forward: dict[str, list[str]] = defaultdict(list)
    all_edge_participants: set[str] = set()

    for role in roles:
        role_name = role.get("name", "")
        for status in role.get("statuses") or []:
            status_type = status.get("roleStatusType")
            edge_type = _STATUS_TYPE_TO_EDGE.get(status_type)
            if not edge_type:
                continue
            for dest in status.get("destinations") or []:
                edges.append(
                    {
                        "from": role_name,
                        "to": dest,
                        "transition_type": edge_type,
                        "status_name": status.get("name"),
                    }
                )
                all_edge_participants.add(role_name)
                all_edge_participants.add(dest)
                if edge_type == "forward":
                    forward_graph[role_name].append(dest)
                    inbound_forward[dest].append(role_name)

    # Normalize role entries using export field names
    # Export keys: name, shortName, type, startRole, businessKey,
    #   visibleForApplicant, assignedTo, institutions[], statuses[]
    # NOT present: id, sortOrderNumber, usedInFlow
    normalized_roles = []
    for idx, role in enumerate(roles):
        role_type = role.get("type", "UserRole")
        role_name = role.get("name", "")

        # Extract institution/unit from first institution entry
        institutions = role.get("institutions") or []
        first_inst = institutions[0] if institutions else {}

        # Normalize statuses for this role
        role_statuses = []
        for status in role.get("statuses") or []:
            edge_type = _STATUS_TYPE_TO_EDGE.get(
                status.get("roleStatusType"), "unknown"
            )
            role_statuses.append(
                {
                    "name": status.get("name", ""),
                    "type": edge_type,
                    "active": status.get("active", True),
                    "destinations": status.get("destinations") or [],
                }
            )

        normalized_roles.append(
            {
                "business_key": role.get("businessKey", ""),
                "name": role_name,
                "short_name": role.get("shortName"),
                "sort_order": idx,
                "role_type": role_type,
                "start_role": role.get("startRole", False),
                "visible_for_applicant": role.get("visibleForApplicant", False),
                "assigned_to": role.get("assignedTo"),
                "used_in_flow": role_name in all_edge_participants,
                "agent_type": _AGENT_TYPE_MAP.get(role_type, "unknown"),
                "institution": first_inst.get("institutionName"),
                "unit": first_inst.get("unitName"),
                "statuses": role_statuses,
            }
        )

    start_roles = [r.get("name", "") for r in roles if r.get("startRole")]

    parallel_groups: list[list[str]] = []
    seen_groups: set[frozenset[str]] = set()
    for _role_name, dests in forward_graph.items():
        if len(dests) > 1:
            group = frozenset(dests)
            if group not in seen_groups:
                seen_groups.add(group)
                parallel_groups.append(sorted(dests))

    convergence_roles = sorted(
        name for name, preds in inbound_forward.items() if len(preds) > 1
    )

    correction_paths = [
        {"from": e["from"], "to": e["to"], "status_name": e["status_name"]}
        for e in edges
        if e["transition_type"] == "corrections"
    ]

    return {
        "roles": normalized_roles,
        "edges": edges,
        "start_roles": start_roles,
        "parallel_groups": parallel_groups,
        "convergence_roles": convergence_roles,
        "correction_paths": correction_paths,
    }


def _build_service_section(
    export_data: dict[str, Any],
    service_data: dict[str, Any],
    instance: str | None,
    warnings: list[str],
) -> dict[str, Any]:
    """Build the service metadata section."""
    updated_at = service_data.get("updatedAt")
    warnings.append(
        "BPA has no distinct publish timestamp; published_at falls back to updated_at"
    )
    return {
        "instance": instance,
        "service_id": str(service_data.get("id", export_data.get("id", ""))),
        "name": service_data.get("name", export_data.get("name", "")),
        "short_name": service_data.get("shortName", export_data.get("shortName")),
        "active": service_data.get("status") == "ACTIVE",
        "source_version": {
            "updated_at": updated_at,
            "published_at": updated_at,
        },
    }


def _build_visibility_map(
    behaviours: list[dict[str, Any]],
    determinants: list[dict[str, Any]],
) -> dict[str, list[dict[str, Any]]]:
    """Build componentKey → [visibility_rule] map from behaviours + determinants.

    Inverts the controller-oriented branch_fields into a target-oriented view:
    for each component, which determinants control its visibility and how.
    """
    # determinant oldId → name lookup
    det_names: dict[str, str] = {}
    for d in determinants:
        old_id = d.get("oldId")
        if old_id:
            det_names[old_id] = d.get("name", "")

    rules: dict[str, list[dict[str, Any]]] = defaultdict(list)
    for behaviour in behaviours:
        comp_key = behaviour.get("componentKey", "")
        if not comp_key:
            continue
        for effect in behaviour.get("effects") or []:
            jd_str = effect.get("jsonDeterminants")
            if not jd_str or jd_str in ("null", "[]"):
                continue
            try:
                parsed = json.loads(jd_str) if isinstance(jd_str, str) else jd_str
            except (json.JSONDecodeError, TypeError):
                continue

            # Collect determinant references
            when: list[str] = []
            for group in parsed:
                for item in group.get("items") or []:
                    det_id = item.get("determinantId")
                    if det_id:
                        when.append(det_names.get(det_id, det_id))

            # Collect property effects (activate/show/disabled)
            props: dict[str, str] = {}
            for pe in effect.get("propertyEffects") or []:
                name = pe.get("name")
                value = pe.get("value")
                if name:
                    props[name] = value

            if when:
                rules[comp_key].append({"when": when, **props})

    return dict(rules)


def _iter_all_components(
    components: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """Recursively collect all components from a FormIO tree."""
    result: list[dict[str, Any]] = []
    for comp in components:
        if not isinstance(comp, dict):
            continue
        result.append(comp)
        for child_key in ("components",):
            children = comp.get(child_key)
            if isinstance(children, list):
                result.extend(_iter_all_components(children))
        columns = comp.get("columns")
        if isinstance(columns, list):
            for col in columns:
                if isinstance(col, dict):
                    col_comps = col.get("components", [])
                    if isinstance(col_comps, list):
                        result.extend(_iter_all_components(col_comps))
    return result


def _build_forms_section(
    data: dict[str, Any],
    include_role_forms: bool,
    warnings: list[str],
) -> dict[str, Any]:
    """Build the forms section from export data."""
    forms: dict[str, Any] = {}

    # --- Applicant form ---
    applicant_page = data.get("applicantFormPage")
    if not isinstance(applicant_page, dict):
        applicant_page = {}

    schema = applicant_page.get("formSchema", {})
    if isinstance(schema, str):
        try:
            schema = json.loads(schema)
        except (json.JSONDecodeError, TypeError):
            schema = {}
    if not isinstance(schema, dict):
        schema = {}

    components = schema.get("components", [])
    if not isinstance(components, list):
        components = []

    all_keys = get_all_component_keys(components) if components else set()
    simplified = simplify_components(components) if components else []

    # Branch analysis using existing helpers
    form_lookup = _build_form_lookup(data)
    determinants = data.get("serviceDeterminants", [])
    behaviours = data.get("componentBehaviours", [])
    branch_result = _build_branch_map(determinants, behaviours, form_lookup)

    # Build per-field visibility rules from behaviours + determinants
    visibility_map = _build_visibility_map(behaviours, determinants)
    for field in simplified:
        key = field.get("key")
        if key and key in visibility_map:
            field["visibility_rules"] = visibility_map[key]

    # Warn about classification-backed fields that couldn't resolve values
    classification_fields = [
        comp.get("key")
        for comp in _iter_all_components(components)
        if comp.get("type") == "select" and comp.get("data", {}).get("dataSrc") == "url"
    ]
    if classification_fields:
        warnings.append(
            f"{len(classification_fields)} classification-backed field(s) could not "
            f"resolve values: {classification_fields}"
        )

    # Warn about components with custom conditional logic
    custom_conditional_keys = [
        comp.get("key")
        for comp in _iter_all_components(components)
        if comp.get("customConditional") or comp.get("logic")
    ]
    if custom_conditional_keys:
        warnings.append(
            f"{len(custom_conditional_keys)} component(s) have custom conditional "
            f"logic not captured in branch_fields: {custom_conditional_keys}"
        )

    # Systematic warnings for contract field coverage
    unresolved_count = sum(
        1 for f in simplified if f.get("value_source") == "unresolved"
    )
    if unresolved_count:
        warnings.append(
            f"{unresolved_count} selectable field(s) have value_source='unresolved' "
            "(values populated at runtime, not in export)"
        )

    forms["applicant"] = {
        "field_count": len(all_keys),
        "branch_fields": branch_result.get("branches", []),
        "fields": simplified,
    }

    # --- Role forms (opt-in) ---
    if include_role_forms:
        roles = data.get("roles", [])
        role_forms: list[dict[str, Any]] = []
        for role in roles:
            role_entry: dict[str, Any] = {
                "role_id": str(role.get("id", "")),
                "role_name": role.get("name", ""),
            }
            role_schema_raw = role.get("formSchema")
            if role_schema_raw:
                role_form_data = {"formSchema": role_schema_raw}
                parsed = parse_form_schema(role_form_data)
                role_components = parsed.get("components", [])
                if isinstance(role_components, list) and role_components:
                    role_keys = get_all_component_keys(role_components)
                    role_entry["field_count"] = len(role_keys)
                    role_entry["fields"] = simplify_components(role_components)
                else:
                    role_entry["field_count"] = 0
                    role_entry["fields"] = []
            else:
                role_entry["field_count"] = None
                role_entry["fields"] = []
                role_entry["availability"] = "no_schema"
                warnings.append(f"Role '{role.get('name')}' has no form schema")
            role_forms.append(role_entry)
        forms["roles"] = role_forms

    return forms


async def _build_documents_section(
    client: Any,
    registrations: list[dict[str, Any]],
    warnings: list[str],
) -> dict[str, Any]:
    """Build documents section by traversing registrations."""
    all_docs: list[dict[str, Any]] = []

    for reg in registrations:
        reg_id = reg.get("id")
        reg_name = reg.get("name", "")
        if not reg_id:
            continue
        try:
            doc_reqs = await client.get_list(
                "/registration/{registration_id}/document_requirement",
                path_params={"registration_id": reg_id},
                resource_type="document_requirement",
            )
            for doc in doc_reqs:
                all_docs.append(
                    {
                        "requirement_id": str(doc.get("id", "")),
                        "name": doc.get("name", ""),
                        "required": bool(doc.get("required", False)),
                        "registration_id": str(reg_id),
                        "registration_name": reg_name,
                    }
                )
        except BPAClientError as e:
            warnings.append(
                f"Failed to fetch documents for registration '{reg_name}': {e}"
            )

    return {"applicant_documents": all_docs}


@large_response_handler(
    threshold_bytes=100 * 1024,  # 100KB threshold
    navigation={
        "flow": "jq '.intended_flow'",
        "applicant_form": "jq '.forms.applicant'",
        "branch_fields": "jq '.forms.applicant.branch_fields'",
        "documents": "jq '.documents'",
        "warnings": "jq '.warnings'",
    },
)
async def service_ai_map(
    service_id: str | int,
    include_role_forms: bool = False,
    include_documents: bool = True,
    include_raw_sources: bool = False,
    instance: str | None = None,
) -> dict[str, Any]:
    """Return a normalized AI-readable service map. Read-only.

    Composes workflow graph, form schemas, branch analysis, and document
    requirements into one deterministic artifact.

    Args:
        service_id: BPA service UUID.
        include_role_forms: Include per-role form schemas (default: False).
        include_documents: Include document requirements (default: True).
        include_raw_sources: Include raw export data (default: False).
        instance: Instance profile name (from instance_list).

    Returns:
        dict with service, intended_flow, forms, documents, warnings, sources.
    """
    if not service_id:
        raise ToolError(
            "'service_id' is required. Use 'service_list' to find valid service IDs."
        )

    warnings: list[str] = []
    warnings.append(
        "Camunda activityId not available from BPA; "
        "use DS service_roles for Camunda IDs"
    )

    try:
        async with BPAClient.for_instance(instance) as client:
            # Verify service exists and get metadata
            try:
                service_data = await client.get(
                    "/service/{id}",
                    path_params={"id": service_id},
                    resource_type="service",
                    resource_id=str(service_id),
                )
            except BPANotFoundError:
                raise ToolError(
                    f"Service '{service_id}' not found. "
                    "Use 'service_list' to see available services."
                )

            # Single export call
            export_data, _ = await client.download_service(
                str(service_id),
                options=_AI_MAP_EXPORT_OPTIONS,
            )

            data = _unwrap_export(export_data)

            # Build sections (client stays open for doc calls in later tasks)
            service_section = _build_service_section(
                data, service_data, instance, warnings
            )

            roles = data.get("roles", [])
            intended_flow = _build_intended_flow(roles, warnings)

            # Registrations summary from export (richer than service_get)
            export_registrations = data.get("registrations") or []
            registrations_summary = [
                {
                    "id": str(reg.get("id", "")),
                    "name": reg.get("name", ""),
                    "short_name": reg.get("shortName"),
                    "business_key": reg.get("businessKey"),
                    "active": reg.get("active", True),
                    "role_names": reg.get("roleNames") or [],
                    "institution": (reg.get("registrationInstitution") or {}).get(
                        "name"
                    ),
                }
                for reg in export_registrations
            ]
            service_section["registrations"] = registrations_summary

            forms_section = _build_forms_section(data, include_role_forms, warnings)

            # Document traversal (opt-in, default True)
            documents_section = None
            if include_documents:
                registrations = service_data.get("registrations", [])
                documents_section = await _build_documents_section(
                    client, registrations, warnings
                )

            result: dict[str, Any] = {
                "service": service_section,
                "intended_flow": intended_flow,
                "forms": forms_section,
                "warnings": warnings,
                "sources": {
                    "export": f"POST /download_service/{service_id}",
                    "export_options": _AI_MAP_EXPORT_OPTIONS,
                },
            }

            if documents_section is not None:
                result["documents"] = documents_section
                result["sources"]["document_requirements"] = (
                    "GET /registration/{id}/document_requirement (per registration)"
                )

            if include_raw_sources:
                result["sources"]["raw_export"] = data

    except ToolError:
        raise
    except BPAClientError as e:
        raise translate_error(e, resource_type="service")

    return result


def register_service_map_tools(mcp: Any) -> None:
    """Register service map tools with the MCP server."""
    from mcp_eregistrations_bpa.tools.annotations import READ

    mcp.tool(annotations=READ)(service_ai_map)



"""DS MCP Server — admin/ops monitoring for eRegistrations Display System."""

from fastmcp import FastMCP

mcp = FastMCP(
    "mcp-eregistrations-ds",
    instructions=(
        "DS (Display System) tools for monitoring and managing application files, "
        "payments, workflows, documents, and system health. "
        "Read tools for monitoring, write tools for file lifecycle "
        "(create, fill, submit, process, pay, cleanup). "
        "Use ds_auth_login to authenticate first."
    ),
)


import mcp_eregistrations_ds.tools.documents  # noqa: F401, E402
import mcp_eregistrations_ds.tools.file_lifecycle  # noqa: F401, E402
import mcp_eregistrations_ds.tools.files  # noqa: F401, E402
import mcp_eregistrations_ds.tools.financial_reports  # noqa: F401, E402
import mcp_eregistrations_ds.tools.messages  # noqa: F401, E402
import mcp_eregistrations_ds.tools.payments  # noqa: F401, E402
import mcp_eregistrations_ds.tools.processes  # noqa: F401, E402
import mcp_eregistrations_ds.tools.services  # noqa: F401, E402
import mcp_eregistrations_ds.tools.system  # noqa: F401, E402
import mcp_eregistrations_ds.tools.users  # noqa: F401, E402
from mcp_eregistrations_common.tools.instances import (  # noqa: E402
    register_instance_tools,
)

register_instance_tools(mcp)


def main() -> None:
    """Entry point for mcp-eregistrations-ds."""
    mcp.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
from typing import Any
from urllib.parse import quote

from fastmcp.exceptions import ToolError
from mcp.types import ToolAnnotations

from mcp_eregistrations_keycloak.audit.logger import KcAuditLogger
from mcp_eregistrations_keycloak.audit.models import KcObjectType
from mcp_eregistrations_keycloak.keycloak_client import (
    KeycloakAdminClient,
    KeycloakClientError,
    translate_error,
)
from mcp_eregistrations_keycloak.server import mcp
from mcp_eregistrations_keycloak.tools.validation import (
    _ROLE_UPDATABLE_FIELDS,
    _extract_items,
    _filter_update_fields,
    _resolve_realm,
    _validate_role_name,
    _validate_user_id,
)

# --- Role Definition Tools ---


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_list_realm_roles(
    realm: str | None = None,
    first: int = 0,
    max: int = 100,
    search: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """List realm-level roles.

    Args:
        realm: Realm name, or None for instance default.
        first: Pagination offset (default: 0).
        max: Page size (default: 100).
        search: Filter by role name substring.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with roles, count.
    """
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            params: dict[str, Any] = {"first": first, "max": max}
            if search:
                params["search"] = search
            roles = await client.get(f"/realms/{r}/roles", **params)
            items = _extract_items(roles)
            return {
                "roles": [
                    {
                        "id": role["id"],
                        "name": role["name"],
                        "description": role.get("description"),
                        "composite": role.get("composite", False),
                    }
                    for role in items
                ],
                "count": len(items),
            }
    except KeycloakClientError as e:
        raise translate_error(e, resource_type="realm role") from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_get_realm_role(
    role_name: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get a realm role by name.

    Args:
        role_name: Role name.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role details.
    """
    _validate_role_name(role_name)
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            path = f"/realms/{r}/roles/{quote(role_name, safe='')}"
            result: dict[str, Any] = await client.get(path)
            return result
    except KeycloakClientError as e:
        raise translate_error(
            e, resource_type="realm role", resource_id=role_name
        ) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_list_role_users(
    role_name: str,
    realm: str | None = None,
    first: int = 0,
    max: int = 20,
    instance: str | None = None,
) -> dict[str, Any]:
    """List users assigned to a realm role.

    Args:
        role_name: Role name.
        realm: Realm name, or None for instance default.
        first: Pagination offset (default: 0).
        max: Page size (default: 20).
        instance: Instance name, or None for auto-select.

    Returns:
        dict with users, role_name.
    """
    _validate_role_name(role_name)
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            users = await client.get(
                f"/realms/{r}/roles/{quote(role_name, safe='')}/users",
                first=first,
                max=max,
            )
            items = _extract_items(users)
            return {
                "role_name": role_name,
                "users": [
                    {
                        "id": u["id"],
                        "username": u["username"],
                        "email": u.get("email"),
                    }
                    for u in items
                ],
            }
    except KeycloakClientError as e:
        raise translate_error(
            e, resource_type="role users", resource_id=role_name
        ) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_list_client_roles(
    client_id: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """List roles for a specific client.

    Args:
        client_id: Keycloak client UUID (not clientId string).
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with roles, client_id.
    """
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            roles = await client.get(f"/realms/{r}/clients/{client_id}/roles")
            items = _extract_items(roles)
            return {
                "client_id": client_id,
                "roles": [
                    {
                        "id": role["id"],
                        "name": role["name"],
                        "description": role.get("description"),
                        "composite": role.get("composite", False),
                    }
                    for role in items
                ],
            }
    except KeycloakClientError as e:
        raise translate_error(
            e, resource_type="client role", resource_id=client_id
        ) from e


@mcp.tool()
async def kc_create_realm_role(
    role_name: str,
    description: str | None = None,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Create a realm role. Audited write operation.

    Args:
        role_name: Unique role name.
        description: Role description.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role_name, status.
    """
    _validate_role_name(role_name)
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None, "create", KcObjectType.ROLE, {"realm": r, "role_name": role_name}
            )
            role_data: dict[str, Any] = {
                "name": role_name,
                "attributes": {"managed-by": ["keycloak-mcp"]},
            }
            if description:
                role_data["description"] = description
            result = await client.post(f"/realms/{r}/roles", data=role_data)
            await logger.save_rollback_state(
                audit_id, KcObjectType.ROLE, role_name, role_data
            )
            outcome = {
                "role_name": role_name,
                "realm": r,
                "status": "created",
                **result,
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e, resource_type="realm role", resource_id=role_name
        ) from e


@mcp.tool()
async def kc_update_realm_role(
    role_name: str,
    updates: dict[str, Any],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Update a realm role. Audited write operation.

    Args:
        role_name: Role name to update.
        updates: Fields to update (name, description, attributes).
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role_name, updated_fields, status.
    """
    _validate_role_name(role_name)
    filtered = _filter_update_fields(updates, _ROLE_UPDATABLE_FIELDS, "role")
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "update",
                KcObjectType.ROLE,
                {"role_name": role_name, "fields": list(filtered.keys())},
            )
            current = await client.get(f"/realms/{r}/roles/{quote(role_name, safe='')}")
            await logger.save_rollback_state(
                audit_id, KcObjectType.ROLE, role_name, current
            )
            merged = {**current, **filtered}
            await client.put(
                f"/realms/{r}/roles/{quote(role_name, safe='')}",
                data=merged,
            )
            outcome = {
                "role_name": role_name,
                "realm": r,
                "updated_fields": list(filtered.keys()),
                "status": "updated",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e, resource_type="realm role", resource_id=role_name
        ) from e


@mcp.tool()
async def kc_delete_realm_role(
    role_name: str,
    force: bool = False,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Delete a realm role. Audited write operation.

    Only allows deletion of MCP-created roles unless force=True.

    Args:
        role_name: Role name to delete.
        force: If True, delete even non-MCP-managed roles (default: False).
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role_name, status.
    """
    _validate_role_name(role_name)
    logger = KcAuditLogger.for_instance(instance)
    audit_id = await logger.record_pending(
        None, "delete", KcObjectType.ROLE, {"role_name": role_name}
    )
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            path = f"/realms/{r}/roles/{quote(role_name, safe='')}"
            if not force:
                role = await client.get(path)
                managed_by = role.get("attributes", {}).get("managed-by", [])
                if "keycloak-mcp" not in managed_by:
                    raise ToolError(
                        f"Role '{role_name}' was not created by MCP. "
                        f"Set force=True to delete non-MCP-managed roles."
                    )
                await logger.save_rollback_state(
                    audit_id, KcObjectType.ROLE, role_name, role
                )
            else:
                role = await client.get(path)
                await logger.save_rollback_state(
                    audit_id, KcObjectType.ROLE, role_name, role
                )
            await client.delete(path)
            outcome = {"role_name": role_name, "realm": r, "status": "deleted"}
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e, resource_type="realm role", resource_id=role_name
        ) from e


# --- Composite Role Tools ---


async def _resolve_roles(
    client: KeycloakAdminClient,
    realm: str,
    roles: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """Resolve role names to full representations (with id) in parallel."""

    async def _resolve_one(role: dict[str, Any]) -> dict[str, Any]:
        if "id" in role:
            return role
        name = role.get("name")
        if not name:
            raise ToolError("Each role must have 'name' or 'id'.")
        resolved: dict[str, Any] = await client.get(
            f"/realms/{realm}/roles/{quote(name, safe='')}"
        )
        return resolved

    return list(await asyncio.gather(*[_resolve_one(r) for r in roles]))


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_get_composite_roles(
    role_name: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get child roles of a composite role.

    Args:
        role_name: Composite role name.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role_name, composites.
    """
    _validate_role_name(role_name)
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            composites = await client.get(
                f"/realms/{r}/roles/{quote(role_name, safe='')}/composites"
            )
            return {
                "role_name": role_name,
                "composites": composites or [],
            }
    except KeycloakClientError as e:
        raise translate_error(
            e,
            resource_type="composite role",
            resource_id=role_name,
        ) from e


@mcp.tool()
async def kc_add_composite_roles(
    role_name: str,
    roles: list[dict[str, Any]],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Add child roles to a composite role. Audited write operation.

    Args:
        role_name: Parent composite role name.
        roles: List of roles to add (each with 'name' or 'id').
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role_name, added_count, status.
    """
    _validate_role_name(role_name)
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "create",
                KcObjectType.COMPOSITE_ROLE,
                {"realm": r, "role_name": role_name},
            )
            resolved = await _resolve_roles(client, r, roles)
            await logger.save_rollback_state(
                audit_id, KcObjectType.COMPOSITE_ROLE, role_name, {"roles": resolved}
            )
            await client.post(
                f"/realms/{r}/roles/{quote(role_name, safe='')}/composites",
                data=resolved,
            )
            outcome = {
                "role_name": role_name,
                "realm": r,
                "added_count": len(resolved),
                "status": "composites_added",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="composite role",
            resource_id=role_name,
        ) from e


@mcp.tool()
async def kc_remove_composite_roles(
    role_name: str,
    roles: list[dict[str, Any]],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Remove child roles from a composite role. Audited write operation.

    Args:
        role_name: Parent composite role name.
        roles: List of roles to remove (each with 'name' or 'id').
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with role_name, removed_count, status.
    """
    _validate_role_name(role_name)
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "delete",
                KcObjectType.COMPOSITE_ROLE,
                {"realm": r, "role_name": role_name},
            )
            resolved = await _resolve_roles(client, r, roles)
            await logger.save_rollback_state(
                audit_id, KcObjectType.COMPOSITE_ROLE, role_name, {"roles": resolved}
            )
            await client.delete(
                f"/realms/{r}/roles/{quote(role_name, safe='')}/composites",
                data=resolved,
            )
            outcome = {
                "role_name": role_name,
                "realm": r,
                "removed_count": len(resolved),
                "status": "composites_removed",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="composite role",
            resource_id=role_name,
        ) from e


# --- User Role Mapping Tools ---


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_get_user_role_mappings(
    user_id: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get all role mappings for a user.

    Args:
        user_id: User UUID.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with user_id, realm_mappings, client_mappings.
    """
    _validate_user_id(user_id)
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            mappings = await client.get(f"/realms/{r}/users/{user_id}/role-mappings")
            return {
                "user_id": user_id,
                "realm_mappings": mappings.get("realmMappings", []),
                "client_mappings": mappings.get("clientMappings", {}),
            }
    except KeycloakClientError as e:
        raise translate_error(
            e,
            resource_type="user role mapping",
            resource_id=user_id,
        ) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_get_effective_realm_roles(
    user_id: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get effective (direct + inherited) realm roles for a user.

    Args:
        user_id: User UUID.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with user_id, roles.
    """
    _validate_user_id(user_id)
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            roles = await client.get(
                f"/realms/{r}/users/{user_id}/role-mappings/realm/composite"
            )
            return {"user_id": user_id, "roles": roles or []}
    except KeycloakClientError as e:
        raise translate_error(
            e,
            resource_type="effective roles",
            resource_id=user_id,
        ) from e


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def kc_get_available_realm_roles(
    user_id: str,
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get realm roles available to assign to a user.

    Args:
        user_id: User UUID.
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with user_id, roles.
    """
    _validate_user_id(user_id)
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            roles = await client.get(
                f"/realms/{r}/users/{user_id}/role-mappings/realm/available"
            )
            return {"user_id": user_id, "roles": roles or []}
    except KeycloakClientError as e:
        raise translate_error(
            e,
            resource_type="available roles",
            resource_id=user_id,
        ) from e


@mcp.tool()
async def kc_assign_realm_roles(
    user_id: str,
    roles: list[dict[str, Any]],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Assign realm roles to a user. Audited write operation.

    Args:
        user_id: User UUID.
        roles: List of roles to assign (each with 'name' or 'id').
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with user_id, assigned_count, status.
    """
    _validate_user_id(user_id)
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "create",
                KcObjectType.USER_ROLE_MAPPING,
                {"realm": r, "user_id": user_id},
            )
            resolved = await _resolve_roles(client, r, roles)
            await logger.save_rollback_state(
                audit_id, KcObjectType.USER_ROLE_MAPPING, user_id, {"roles": resolved}
            )
            await client.post(
                f"/realms/{r}/users/{user_id}/role-mappings/realm",
                data=resolved,
            )
            outcome = {
                "user_id": user_id,
                "realm": r,
                "assigned_count": len(resolved),
                "status": "roles_assigned",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="user role assignment",
            resource_id=user_id,
        ) from e


@mcp.tool()
async def kc_remove_user_realm_roles(
    user_id: str,
    roles: list[dict[str, Any]],
    realm: str | None = None,
    instance: str | None = None,
) -> dict[str, Any]:
    """Remove realm roles from a user. Audited write operation.

    Args:
        user_id: User UUID.
        roles: List of roles to remove (each with 'name' or 'id').
        realm: Realm name, or None for instance default.
        instance: Instance name, or None for auto-select.

    Returns:
        dict with user_id, removed_count, status.
    """
    _validate_user_id(user_id)
    logger = KcAuditLogger.for_instance(instance)
    audit_id: str | None = None
    try:
        async with KeycloakAdminClient.for_instance(instance) as client:
            r = _resolve_realm(realm, instance, client)
            audit_id = await logger.record_pending(
                None,
                "delete",
                KcObjectType.USER_ROLE_MAPPING,
                {"realm": r, "user_id": user_id},
            )
            resolved = await _resolve_roles(client, r, roles)
            await logger.save_rollback_state(
                audit_id, KcObjectType.USER_ROLE_MAPPING, user_id, {"roles": resolved}
            )
            await client.delete(
                f"/realms/{r}/users/{user_id}/role-mappings/realm",
                data=resolved,
            )
            outcome = {
                "user_id": user_id,
                "realm": r,
                "removed_count": len(resolved),
                "status": "roles_removed",
            }
            await logger.mark_success(audit_id, outcome)
            return outcome
    except KeycloakClientError as e:
        if audit_id is not None:
            await logger.mark_failed(audit_id, str(e))
        raise translate_error(
            e,
            resource_type="user role mapping",
            resource_id=user_id,
        ) from e



"""Graylog stream tools — list and inspect streams."""

from __future__ import annotations

from typing import Any

from mcp.types import ToolAnnotations

from mcp_eregistrations_graylog.graylog_client import (
    GraylogClient,
    GraylogClientError,
    translate_error,
)
from mcp_eregistrations_graylog.server import mcp


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def graylog_stream_list(
    instance: str | None = None,
) -> dict[str, Any]:
    """List all Graylog streams. Requires auth.

    Args:
        instance: Instance name.

    Returns:
        dict with count, streams (id, title, description, disabled).
    """
    try:
        async with GraylogClient.for_instance(instance) as client:
            data = await client.get("/api/streams")
    except GraylogClientError as e:
        raise translate_error(e, resource_type="streams") from e

    raw_streams = data.get("streams", [])
    streams = [
        {
            "id": s.get("id"),
            "title": s.get("title"),
            "description": s.get("description"),
            "disabled": s.get("disabled", False),
        }
        for s in raw_streams
    ]
    return {"count": len(streams), "streams": streams}


@mcp.tool(annotations=ToolAnnotations(readOnlyHint=True))
async def graylog_stream_get(
    stream_id: str,
    instance: str | None = None,
) -> dict[str, Any]:
    """Get detailed stream info. Requires auth.

    Args:
        stream_id: Stream ID (from graylog_stream_list).
        instance: Instance name.

    Returns:
        dict with stream details (id, title, rules, created_at, etc).
    """
    try:
        async with GraylogClient.for_instance(instance) as client:
            result: dict[str, Any] = await client.get(f"/api/streams/{stream_id}")
            return result
    except GraylogClientError as e:
        raise translate_error(e, resource_type="stream", resource_id=stream_id) from e
